<?php
// tòmò web analytics
// Copyright since 2011 by Yves Goergen
// Website: http://unclassified.de/go/tomo
// Licence: GNU General Public License (GPL) Version 3
// Encoding: UTF-8 (without BOM) (auto-detect: °°°°°)
//
// This file is derived from:
// UNB Components
// Copyright since 2007 by Yves Goergen
// Website: http://unclassified.de/go/unb2
// Licence: GNU General Public License (GPL) Version 3
//
// This file contains the UNB locale base class.

// Contains format flags for localised formatted dates and times.
//
class UnbTimeFormat
{
	// Short day of week.
	const Weekday = 1;
	// Long day of week.
	const WeekdayLong = 2;
	// Short numeric date.
	const Date = 4;
	// Long date.
	const DateLong = 8;
	// Time of day with seconds.
	const TimeSec = 16;
	// Time of day without seconds.
	const Time = 32;
	// Time zone abbreviation.
	const TimeZone = 64;

	// Standard format for date and time.
	const DateTime = 36;
	// Standard format for date and time with seconds.
	const DateTimeSec = 20;
	// Standard format for long date and time.
	const DateTimeLong = 40;
	// Standard format for long date and time with seconds.
	const DateTimeSecLong = 24;
} // class UnbTimeFormat

// Enumeration of relative time grammatical contexts.
//
class UnbTimeContext
{
	// The time describes a point in time in the past or future.
	const Point = 1;
	// The time describes a time span since or until now.
	const Span = 2;
	// The time describes a point in time before another point in time in the past.
	// This is used in quotations.
	const Before = 3;
} // class UnbTimeContext

// UNB localisation class.
//
class UnbLocale
{
	// Locale code, to be set by sub-classes.
	protected $code = null;

	// Locale importance for its language, used to sort known locales when only a language is requested.
	// This should be roughly the number of people affected by the locale, divided by 1000.
	protected $importance = 0;

	// Locale code that the locale depends on for dictionary translations. To be set by sub-classes
	// that contain an incomplete dictionary and need a certain other locale available as fallback,
	// like for example en-GB only includes deviations from en-US and otherwise needs to fall back
	// to en-US.
	protected $dependsOn = null;

	// The text dictionary, to be set by sub-classes.
	protected $dict = array();

	// Starting quotation mark.
	protected $quoteStart = '"';

	// Ending quotation mark.
	protected $quoteEnd = '"';

	// Starting parentheses character.
	protected $parentheseStart = '(';

	// Ending parentheses character.
	protected $parentheseEnd = ')';

	// Colon.
	protected $colon = ':';

	// Known available locale objects.
	private static $knownLocales = array();

	// Preferred locale objects to use, in this order.
	private static $prefLocales = array();

	// (DateTimeZone) Current timezone. If unset, the server-default will be used.
	private static $timezone = null;

	// Automatically loads and registers all locale classes in a directory.
	//
	// All locale class files must be named like *.locale.class.php. The locale code does not need
	// to be included in the file name in a special way as the class files will register
	// themselves upon inclusion by PHP.
	//
	// This function is called automatically when AutoPrefer is called and no locales have been
	// registered yet.
	//
	// path = (string) Locale classes directory. If unset, the directory 'locales' under this code file is used.
	//
	public static function AutoRegister($path = null)
	{
		if (!isset($path))
			$path = dirname(__FILE__) . '/locales';
		if (!is_string($path))
			throw new InvalidArgumentException('Invalid argument type: $path');

		foreach (glob($path . '/*.locale.class.php') as $file)
		{
			include_once($file);
		}
	}

	// Registers a new locale object.
	//
	// This function is used by the code files of locale classes to have the defined locale class
	// automatically registered.
	//
	// obj = (UnbLocale) New locale instance to be registered
	//
	public static function Register($obj)
	{
		if (!($obj instanceof UnbLocale))
			throw new InvalidArgumentException('Invalid argument type: $obj');

		self::$knownLocales[$obj->code] = $obj;
	}

	// Prefer the specified locale.
	//
	// The locales are preferred in the order in that they have been passed to this function, the
	// first added being regarded first, i.e. having the highest priority.
	//
	// If no locales have been registered yet, the function AutoRegister is called to find all
	// locale classes in a directory.
	//
	// code = (string) Locale code
	// clientSorted = (bool) {{true}} to sort language-only by client accepted language,
	//                       {{false}} to sort by locale-provided importance.
	//                       (Only applies to language-only (two-letter) codes.)
	//
	public static function Prefer($code, $clientSorted = false)
	{
		if (!is_string($code))
			throw new InvalidArgumentException('Invalid argument type: $code');
		if (strlen($code) != 2 && strlen($code) != 5)
			throw new InvalidArgumentException('Argument out of range: $code');
		if (!is_bool($clientSorted))
			throw new InvalidArgumentException('Invalid argument type: $clientSorted');

		if (!count(self::$knownLocales))
			self::AutoRegister();

		if (!count(self::$knownLocales))
			throw new Exception('No locales defined');

		if (strlen($code) == 2)
		{
			// Only the language is specified, no country/region.
			// Prefer all known locales of that language, sorted by their importance.
			$codes = array();
			$sort = array();
			foreach (self::$knownLocales as $kl)
			{
				if (substr($kl->code, 0, 2) == $code)
				{
					$codes[] = $kl->code;
					$sort[] = $kl->importance;
				}
			}
			if (!$clientSorted)
			{
				array_multisort($sort, SORT_DESC, $codes);
				foreach ($codes as $code2)
				{
					self::Prefer($code2);
				}
			}
			else
			{
				foreach (UnbEnvironment::GetClientLanguages() as $clientCode)
				{
					if (in_array($clientCode, $codes))
						self::Prefer($clientCode);
				}
			}
		}
		else
		{
			// Full locale identifier
			foreach (self::$prefLocales as $pl)
			{
				if ($pl->code == $code) return;   // Already set, nothing to do
			}
			foreach (self::$knownLocales as $kl)
			{
				if ($kl->code == $code)
				{
					self::$prefLocales[] = $kl;
					#echo '### Preferring ' . $code . ' ### ';
					if (isset($kl->dependsOn))
					{
						// The locale dictionary is incomplete and its dependent locale must be
						// preferred after it as fallback.
						self::Prefer($kl->dependsOn);
					}
					return;
				}
			}
		}
	}

	// Automatically determines the locales to prefer from the user agent.
	//
	// If no locales have been registered yet, the function AutoRegister is called to find all
	// locale classes in a directory.
	//
	public static function AutoPrefer()
	{
		if (!count(self::$knownLocales))
			self::AutoRegister();

		if (!count(self::$knownLocales))
			throw new Exception('No locales defined');

		self::$prefLocales = array();
		// Prefer the language of the page we're serving, if set
		$pageLanguage = null; #UnbEnvironment::PageLanguage();
		if (isset($pageLanguage))
		{
			// For the URL-defined language, prefer all locales of this language in the order of client preference.
			self::Prefer($pageLanguage, true);
		}
		// Prefer the locales as defined by the client
		foreach (UnbEnvironment::GetClientLanguages() as $code)
		{
			self::Prefer($code);
		}
		// If none matched, prefer en-US
		if (!count(self::$prefLocales))
		{
			self::Prefer('en-US');
		}
		// If this doesn't exist, prefer whatever is available
		if (!count(self::$prefLocales) && count(self::$knownLocales))
		{
			self::Prefer(self::$knownLocales[0]);
		}
		// If no locales are available, you will see an exception soon
	}

	// Gets the code of the currently most-preferred locale.
	//
	// Returns (string) The most-preferred locale code, or {{null}} if no locale is selected.
	//
	public static function GetCode()
	{
		if (count(self::$prefLocales))
		{
			return self::$prefLocales[0]->code;
		}
		return null;
	}

	// Gets all preferred locale codes, in their order of preference.
	//
	// Returns (array(string))
	//
	public static function GetPreferredLocaleCodes()
	{
		$a = array();
		foreach (self::$prefLocales as $pl)
		{
			$a[] = $pl->code;
		}
		return $a;
	}

	// Sets the default time zone to display time specifications in.
	//
	// timezone = (DateTimeZone) PHP DateTimeZone object
	//            (string) Time zone name as included in the tz database.
	//
	public static function SetTimeZone($timezone)
	{
		if (is_string($timezone))
			$timezone = new DateTimeZone($timezone);
		else if (!($obj instanceof DateTimeZone))
			throw new InvalidArgumentException('Invalid argument type: $timezone');

		self::$timezone = $timezone;
	}

	// Adds more words to a locale's dictionary.
	//
	// code = (string) Registered locale code
	// dict = (array(key => text)) Dictionary entries, as defined in the locale
	//
	public static function AddDict($code, $dict)
	{
		if (!is_string($code))
			throw new InvalidArgumentException('Invalid argument type: $code');
		if (!is_array($dict))
			throw new InvalidArgumentException('Invalid argument type: $dict');

		if (!count(self::$knownLocales))
			self::AutoRegister();

		$locale = self::$knownLocales[$code];
		if (!isset($locale))
			return;

		$locale->dict = array_merge($locale->dict, $dict);
	}

	// Gets the translation of a text.
	//
	// This function has some hidden parameters. The first parameter is always the text key.
	//
	// Then follows the data to be inserted in the variable slots in the translated text. Specify
	// each variable by two parameters: first the variable name, then the value.
	//
	// If the number of parameters after the text key is odd (i.e. there's no exact variable/value
	// match), then the first parameter is taken for a numeric selector of the translated text. It
	// can be used to get different translations for one or multiple entities. If such a number is
	// given, the special variable {#} is resolved to it.
	//
	// If there is only one additional parameter and it is {{true}}, then the text is translated
	// in "safe mode": if no translated text can be found in the dictionary, any namespace is
	// stripped from the text key and the resulting key is returned as translated text. This can
	// be used to have certain abbreviations or terms translated optionally, but use the original
	// term as default. An example is time zone abbreviations, of which every locale is likely to
	// translate only a few.
	//
	// key = (string) Text translation key
	//
	// Returns: (string) The translated text with all specified variables replaced, or {{null}} if
	// no translation is specified in the dictionary (unless "safe mode" is enabled, see above).
	//
	public static function Translate($key)
	{
		if (!is_string($key))
			throw new InvalidArgumentException('Invalid argument type: $key');

		// Get all additional function arguments
		$args = array_slice(func_get_args(), 1);

		// If the only additional parameter is true, enable safe mode
		if (count($args) == 1 && $args[0] === true)
		{
			$safe = true;
			array_shift($args);
		}

		// If an odd number of additional arguments has been specified, take the first one as number parameter
		if (count($args) % 2)
		{
			$num = intval(array_shift($args));
		}

		// Determine the requested key prefix
		if (($pos = strrpos($key, ':')) !== false)
		{
			$prefix = substr($key, 0, $pos);
		}

		// Find the text in one of the selected locales
		foreach (self::$prefLocales as $pl)
		{
			// Check whether the current locale doesn't want to have keys with this prefix
			// translated at all. This can be used when the key already contains a word or name in
			// the language of this locale that should be translated for other locales but not for
			// this one. For example, for keys like 'city:Nuremberg' and 'city:Lisbon', the en-US
			// locale could define a key 'city:*' to false in order to keep the original English
			// city names but have them translated for other locales that define them, like in
			// de-DE 'city:Nuremberg' => 'Nürnberg' or in pt-PT 'city:Lisbon' => 'Lisboa'.
			if ($pl->dict[$prefix . ':*'] === false) break;

			if (isset($num))
			{
				$s = $pl->dict[$key . '#' . $num];
				if (!isset($s))
					$s = $pl->dict[$key . '#' . ($num % 100) . '%100'];
				if (!isset($s))
					$s = $pl->dict[$key . '#' . ($num % 10) . '%10'];
				if (!isset($s))
					$s = $pl->dict[$key];
			}
			else
			{
				$s = $pl->dict[$key];
			}

			if (isset($s)) break;

			// Check whether the current locale doesn't want to have keys with this prefix
			// translated by other locales even if it doesn't provide a translation itself. This
			// can be used when some of the keys are already correct in some languages and every
			// locale defines translations for a different subset of all keys. For example, for
			// keys like 'city:Nuremberg' and 'city:Lisboa', the de-DE locale could define the
			// translation 'city:Nuremberg' => 'Nürnberg' and 'city:Lisboa' => 'Lissabon', where
			// the en-US locale only translates 'city:Lisboa' => 'Lisbon' and the pt-PT locale
			// contains none of these. Then each locale can set 'city:#nofallback' => true to
			// avoid a user prefering the locales pt-PT and then de-DE to get the German name
			// for Lisbon where the Portuguese name is more appropriate, only because the pt-PT
			// dictionary doesn't set a name.
			if (!isset($s) && $pl->dict[$prefix . ':#nofallback'] === true) break;
		}

		if (!isset($s))
		{
			if ($safe)
			{
				// Cut off key namespace and return the remaining key name as untranslated text
				if (($pos = strrpos($key, ':')) !== false)
				{
					$key = substr($key, $pos + 1);
				}
				return $key;
			}
			return null;
		}

		// Insert data
		if (isset($num))
		{
			$s = str_replace('{#}', $num, $s);
		}
		for ($pos = 0; $pos + 1 < sizeof($args); $pos += 2)
		{
			$s = str_replace('{' . $args[$pos] . '}', $args[$pos + 1], $s);
		}

		return $s;
	}

	// Gets all dictionary entries (partially) matching the search key.
	//
	// search = (string) Key or partial key to search for
	//
	// Returns (array(key => text)) Matching dictionary entries
	//
	public static function GetAllTexts($search)
	{
		if (!is_string($search))
			throw new InvalidArgumentException('Invalid argument type: $search');

		$texts = array();
		foreach (self::$prefLocales as $pl)
		{
			foreach ($pl->dict as $key => $text)
			{
				if (!isset($texts[$key]) && strpos($key, $search) !== false)
				{
					$texts[$key] = $text;
				}
			}
		}
		return $texts;
	}

	// Translates a file content type description.
	//
	// contentType = (string) Content type to get the description of
	//
	// Returns (string) Content type description.
	//
	public static function ContentType($contentType)
	{
		if (!is_string($contentType))
			throw new InvalidArgumentException('Invalid argument type: $contentType');

		$str = self::Translate('contenttype:' . $contentType);
		#if (!isset($str) && preg_match('~^(.+?)/~', $contentType, $m))
		#{
		#	$str = self::Translate('contenttype:' . $m[1]);
		#}
		if (!isset($str))
		{
			$str = UnbMessage::GetContentTypeDescription($contentType);
		}
		if (!isset($str))
		{
			$str = self::Translate('contenttype:unknown');
		}
		return $str;
	}

	// Converts the first letter to upper case.
	//
	// str = (string) String to convert
	//
	// Returns (string) String with first letter in upper case.
	//
	public static function Capital($str)
	{
		if (!isset($str)) return null;
		if (!is_string($str))
			throw new InvalidArgumentException('Invalid argument type: $str');

		return mb_strtoupper(mb_substr($str, 0, 1)) . mb_substr($str, 1);
	}

	// Puts the specified string in quotation marks.
	//
	// Returns (string) String with quotation marks.
	//
	public static function Quote($str)
	{
		if (!is_string($str))
			throw new InvalidArgumentException('Invalid argument type: $str');

		if (count(self::$prefLocales))
		{
			return self::$prefLocales[0]->quoteStart .
				$str .
				self::$prefLocales[0]->quoteEnd;
		}
		return '"' . $str . '"';
	}

	// Puts the specified string in parentheses.
	//
	// Returns (string) String with parentheses.
	//
	public static function Parentheses($str)
	{
		if (!is_string($str))
			throw new InvalidArgumentException('Invalid argument type: $str');

		if (count(self::$prefLocales))
		{
			return self::$prefLocales[0]->parentheseStart .
				$str .
				self::$prefLocales[0]->parentheseEnd;
		}
		return '(' . $str . ')';
	}

	// Gets the localised colon.
	//
	// Returns (string) Localised colon.
	//
	public static function Colon()
	{
		if (count(self::$prefLocales))
		{
			return self::$prefLocales[0]->colon;
		}
		return ':';
	}

	// Formats a number.
	//
	// number = (int/float/numeric) Number to format
	// decimals = (int) Number of decimal digits
	// html = (bool) Prepare for HTML output
	// trim = (bool) Remove trailing decimal digits if they are 0
	//
	// Returns (string) Formatted number
	//
	public static function Number($number, $decimals = 0, $html = false, $trim = true)
	{
		if (!is_int($number) && !is_float($number) && !is_numeric($number))
			throw new InvalidArgumentException('Invalid argument type: $number');
		if (!is_int($decimals))
			throw new InvalidArgumentException('Invalid argument type: $decimals');
		if (!is_bool($html))
			throw new InvalidArgumentException('Invalid argument type: $html');
		if (!is_bool($trim))
			throw new InvalidArgumentException('Invalid argument type: $trim');

		if (count(self::$prefLocales))
		{
			return self::$prefLocales[0]->FormatNumber($number, $decimals, $html, $trim);
		}
		return $number;
	}

	// Parses a numeric value from a string.
	//
	// str = (string) Numeric value string to parse.
	// defaultValue = (int/float) Default value to return for empty strings. If unset, an exception may be thrown.
	// recoveryValue = (int/float) Value to return for invalid strings. If unset, an exception may be thrown.
	//
	// Returns (int/float) Number value
	//
	public static function ParseNumber($str, $defaultValue = null, $recoveryValue = null)
	{
		if (!is_string($str))
			throw new InvalidArgumentException('Invalid argument type: $str');
		if (isset($defaultValue))
		{
			if (!is_int($defaultValue) && !is_float($defaultValue))
				throw new InvalidArgumentException('Invalid argument type: $defaultValue');
		}
		if (isset($recoveryValue))
		{
			if (!is_int($recoveryValue) && !is_float($recoveryValue))
				throw new InvalidArgumentException('Invalid argument type: $recoveryValue');
		}
		if (!strlen($str))
		{
			if (isset($defaultValue))
				return $defaultValue;
			throw new InvalidArgumentException('No value specified.');
		}

		if (count(self::$prefLocales))
		{
			return self::$prefLocales[0]->DoParseNumber($str, $recoveryValue);
		}
		if (is_numeric($str))
		{
			if (strpos($str, '.') !== false)
				return floatval($str);
			else
				return intval($str);
		}
		throw new InvalidArgumentException('Invalid numeric value format.');
	}

	// Formats an ordinal number.
	//
	// number = (int) Number to format
	// html = (bool) Prepare for HTML output
	//
	// Returns (string) Formatted number
	//
	public static function OrdinalNumber($number, $html = false)
	{
		if (!is_int($number))
			throw new InvalidArgumentException('Invalid argument type: $number');
		if (!is_bool($html))
			throw new InvalidArgumentException('Invalid argument type: $html');

		if (count(self::$prefLocales))
		{
			return self::$prefLocales[0]->FormatOrdinalNumber($number, $html);
		}
		return $number . '.';
	}

	// Formats a data or file size in bytes.
	//
	// size = (int) Data or file size in bytes
	// html = (bool) Prepare for HTML output
	//
	// Returns (string) Formatted size
	//
	public static function DataSize($size, $html = false)
	{
		if (!is_int($size))
			throw new InvalidArgumentException('Invalid argument type: $size');
		if (!is_bool($html))
			throw new InvalidArgumentException('Invalid argument type: $html');

		if ($size == 0)
			return self::Translate('empty');
		if ($size < 0.9 * 1024)
			return self::NoBreak(self::Translate('unit:bytes', $size));
		if ($size < 50 * 1024)
			return self::NoBreak(self::Number($size / 1024, 1, $html) . ' KiB');
		if ($size < 0.9 * 1024 * 1024)
			return self::NoBreak(self::Number($size / 1024, 0, $html) . ' KiB');
		if ($size < 50 * 1024 * 1024)
			return self::NoBreak(self::Number($size / 1024 / 1024, 1, $html) . ' MiB');
		if ($size < 0.9 * 1024 * 1024 * 1024)
			return self::NoBreak(self::Number($size / 1024 / 1024, 0, $html) . ' MiB');
		if ($size < 50 * 1024 * 1024 * 1024)
			return self::NoBreak(self::Number($size / 1024 / 1024 / 1024, 1, $html) . ' GiB');
		return self::NoBreak(self::Number($size / 1024 / 1024 / 1024, 0, $html) . ' GiB');
	}

	// Formats an absolute time description.
	//
	// time = (string) Time specification
	//        (int) UNIX timestamp
	// details = (UnbTimeFormat) Details of the time specification to include in the output
	// html = (bool) Prepare for HTML output
	// utc = (bool) Treat the specified times as UTC and display it so, i.e. do not convert it to local time.
	//
	// Returns (string) Formatted time
	//
	public static function Time($time, $details = UnbTimeFormat::DateTime, $html = false, $utc = false)
	{
		if (!is_int($time) && !is_string($time))
			throw new InvalidArgumentException('Invalid argument type: $time');
		if (!is_int($details))
			throw new InvalidArgumentException('Invalid argument type: $details');
		if ($details <= 0)
			throw new InvalidArgumentException('Argument out of range: $details');
		if (!is_bool($html))
			throw new InvalidArgumentException('Invalid argument type: $html');

		if (!count(self::$prefLocales))
			throw new BadMethodCallException('No locale selected');

		if (!isset($time) ||
			$time === '' ||
			$time === 0)
		{
			return self::Translate('time:none');
		}

		if (is_int($time) || is_numeric($time))
			$time = '@' . $time;

		$dt = new DateTime($time, new DateTimeZone('UTC'));
		if (!$utc)
		{
			if (isset(self::$timezone))
			{
				$dt->setTimeZone(self::$timezone);
			}
			else
			{
				// Get default timezone from server configuation
				$d = new DateTime('now');
				$tz = $d->getTimeZone();
				$dt->setTimeZone($tz);
			}
		}

		return self::$prefLocales[0]->FormatTime($dt, $details, $html);
	}

	// Formats a relative time description.
	//
	// time = (string) Time specification
	//        (int) UNIX timestamp
	// context = (UnbTimeContext) Grammatical context
	// html = (bool) Prepare for HTML output
	//
	// Returns (string) Formatted time
	//
	public static function RelativeTime($time, $context = UnbTimeContext::Point, $html = false)
	{
		if (!is_int($time) && !is_string($time))
			throw new InvalidArgumentException('Invalid argument type: $time');
		if (!is_int($context))
			throw new InvalidArgumentException('Invalid argument type: $context');
		if ($context <= 0)
			throw new InvalidArgumentException('Argument out of range: $context');
		if (!is_bool($html))
			throw new InvalidArgumentException('Invalid argument type: $html');

		if (!count(self::$prefLocales))
			throw new BadMethodCallException('No locale selected');

		if (!isset($time) ||
			$time === '' ||
			$time === 0)
		{
			return self::Translate('time:never');
		}

		if (is_int($time) || is_numeric($time))
			$time = '@' . $time;

		// DEBUG
		#$time = '2011-02-17 20:11:51';

		$utc = new DateTimeZone('UTC');
		$dt = new DateTime($time, $utc);
		$now = new DateTime('now', $utc);
		if ($dt == $now)
		{
			return self::Translate('time:now');
		}

		$diff = $now->diff($dt);

		// DEBUG
		#$diff = new DateInterval('PT2M24S');
		#$diff = new DateInterval('P1YT2H24S');
		#$diff = new DateInterval('P2Y6M');
		#$diff = new DateInterval('PT23H48M24S');
		#$diff->invert = true;
		#$context = UnbTimeContext::Point;

		#echo $diff->format('PT%hH%iM%sS ');
		#echo $diff->format('P%yY%mM%dDT%hH%iM%sS ');

		// Snap to grid
		// Only snap a unit downwards if there is a greater unit.
		if (($diff->y || $diff->m || $diff->d || $diff->h || $diff->i) && $diff->s <= 3)
		{
			$diff->s = 0;
		}
		else if ($diff->s >= 57)
		{
			$diff->i = $diff->i + 1;   // ++ has no effect here!
			$diff->s = 0;
		}
		if (($diff->y || $diff->m || $diff->d || $diff->h) && $diff->i <= 3)
		{
			$diff->i = 0;
		}
		else if ($diff->i >= 57)
		{
			$diff->h = $diff->h + 1;   // ++ has no effect here!
			$diff->i = 0;
		}
		if (($diff->y || $diff->m || $diff->d) && $diff->h <= 1)
		{
			$diff->h = 0;
		}
		else if ($diff->h >= 23)
		{
			$diff->d = $diff->d + 1;   // ++ has no effect here!
			$diff->h = 0;
		}
		if (($diff->y || $diff->m) && $diff->d <= 2)
		{
			$diff->d = 0;
		}
		else if ($diff->d >= 28)
		{
			$diff->m = $diff->m + 1;   // ++ has no effect here!
			$diff->d = 0;
		}
		// (Months are not snapped)

		// DEBUG
		#echo $diff->format('P%yY%mM%dDT%hH%iM%sS ');

		// Decrease resolution over time, also ensure that only one smaller unit is set
		if ($diff->y >= 5)
		{
			if ($diff->m >= 7)
				$diff->y = $diff->y + 1;   // ++ has no effect here!
			$diff->m = 0;
		}
		if ($diff->y >= 1)
		{
			$diff->d = 0;
			$diff->h = 0;
			$diff->i = 0;
			$diff->s = 0;
		}
		if ($diff->m >= 5)
		{
			if ($diff->d >= 16)
				$diff->m = $diff->m + 1;   // ++ has no effect here!
			$diff->d = 0;
		}
		if ($diff->m >= 1)
		{
			$diff->h = 0;
			$diff->i = 0;
			$diff->s = 0;
		}
		if ($diff->d >= 5)
		{
			if ($diff->h >= 12)
				$diff->d = $diff->d + 1;   // ++ has no effect here!
			$diff->h = 0;
		}
		if ($diff->d >= 1)
		{
			$diff->i = 0;
			$diff->s = 0;
		}
		if ($diff->h >= 5)
		{
			if ($diff->i >= 30)
				$diff->h = $diff->h + 1;   // ++ has no effect here!
			$diff->i = 0;
		}
		if ($diff->h >= 1)
		{
			$diff->s = 0;
		}
		if ($diff->i >= 5)
		{
			if ($diff->s >= 30)
				$diff->i = $diff->i + 1;   // ++ has no effect here!
			$diff->s = 0;
		}

		// Let the locale do the formulation
		return self::$prefLocales[0]->FormatRelativeTime($diff, $context, $html);
	}

	// Prevents a line break between digits and the following text.
	//
	// str = (string) String to be processed
	//
	// Returns (string) Processed string
	//
	public static function NoBreak($str)
	{
		if (!is_string($str))
			throw new InvalidArgumentException('Invalid argument type: $str');

		return preg_replace('~(\d) ~', "\\1\xC2\xA0", $str);   // "\xC2\xA0" = utf8_encode("\xA0")
	}

	protected function FormatNumber($number, $decimals, $html)
	{
		throw new BadMethodCallException('Method not implemented in locale.');
	}

	protected function DoParseNumber($str, $recoveryValue)
	{
		throw new BadMethodCallException('Method not implemented in locale.');
	}

	protected function FormatOrdinalNumber($number, $html)
	{
		throw new BadMethodCallException('Method not implemented in locale.');
	}

	protected function FormatTime($datetime, $details, $html)
	{
		throw new BadMethodCallException('Method not implemented in locale.');
	}

	protected function FormatRelativeTime($diff, $context, $html)
	{
		throw new BadMethodCallException('Method not implemented in locale.');
	}

} // class UnbLocale

?>
