<?php
// tòmò web analytics
// Copyright since 2011 by Yves Goergen
// Website: http://unclassified.de/go/tomo
// Licence: GNU General Public License (GPL) Version 3
// Encoding: UTF-8 (without BOM) (auto-detect: °°°°°)
//
// This file contains the Tomo class.

#if (!class_exists('UnbDatabase', false)) require_once dirname(__FILE__) . '/database.class.php';

// Tomo core functionality.
//
class Tomo
{
	// Lock ID
	private $lockId;
	// New visitor ID to send in a cookie to the client. No cookie is sent if unset.
	private $newVisitorId;
	// Data to send in the JSON response to the client.
	private $response = array();
	// Level of acknowledgement of the DNT header. This is only set if the DNT=1 header is set.
	// It remains 0 otherwise.
	private $trackDntLevel = 0;
	
	private $locationValuesCache;
	private $referrerValuesCache;
	private $browserValuesCache;
	
	private $startTime;
	private $dnsTime;
	private $lockWaitTime;
	
	private $writePostLog = false;
	private $crossDomainSupport = true;
	
	// ===================================  DATA POST FUNCTIONS  ==================================
	

	// Shortened data keys (must be synchronised with the client-side script)
	const key_pageviewId        = 'pv';
	const key_visitorId         = 'vi';
	const key_newVisitor        = 'nv';
	const key_newSession        = 'ns';
	const key_pageTitle         = 'pt';
	const key_pageLanguage      = 'la';
	const key_utcOffset         = 'ut';
	const key_screenWidth       = 'sw';
	const key_screenHeight      = 'sh';
	const key_screenAvailWidth  = 'aw';
	const key_screenAvailHeight = 'ah';
	const key_screenColors      = 'sc';
	const key_cookies           = 'co';
	const key_pageLoadTime      = 'pl';
	const key_campaign          = 'ca';
	const key_pageUrl           = 'pu';
	const key_referrer          = 're';
	const key_viewportWidth     = 'vw';
	const key_viewportHeight    = 'vh';
	const key_flash             = 'fl';
	const key_java              = 'ja';
	const key_quicktime         = 'qt';
	const key_silverlight       = 'sl';
	const key_windowsmedia      = 'wm';
	const key_adblock           = 'ab';
	const key_prerender         = 'pr';
	const key_origPageviewId    = 'op';
	const key_activityTimespan  = 'at';
	const key_duplicatePageview = 'dp';
	// Action data keys
	const key_outlinkUrl        = 'ou';
	const key_downloadUrl       = 'du';
	const key_imgZoomUrl        = 'iu';
	const key_sectionExpand     = 'se';
	const key_siteSearch        = 'ss';
	// Variable prefix keys
	// Visitor variable:  vv_
	// Session variable:  sv_
	// Pageview variable: pv_
	
	// Page action types
	const action_Outlink = 1;
	const action_Download = 2;
	const action_ImgZoom = 3;
	const action_SectionExpand = 4;
	const action_SiteSearch = 5;

	const varNamePattern = '/^[a-z][a-z0-9]*$/i';

	// Internet Explorer XDomainRequest fixup:
	// The content type is set to text/plain so PHP won't parse the POSTed data into $_POST and we
	// have to do it ourselves now. An empty $_POST array is the only indicator, so let's just get the
	// data and peek into it to see whether we find the magic cookie "!POST&". Only then, the data is
	// parsed into the $_POST array fur further use.
	public static function FixParsePostData()
	{
		if (empty($_POST))
		{
			$dataStr = file_get_contents('php://input');
			if (substr($dataStr, 0, 6) == '!POST&')
			{
				parse_str(substr($dataStr, 6), $_POST);
				$_POST['_fromText'] = true;   // DEBUG info only
				$_REQUEST = array_merge($_GET, $_POST);
			}
		}
	}

	// Handles a tracker request.
	//
	// A tracker request is initiated by the JavaScript tracker code or an image request from the
	// browser. It can indicate a pageview or an action on a page. All submitted data is evaluated
	// and stored into the database in this method.
	//
	public function HandlePostRequest()
	{
		$this->startTime = microtime(true);
		
		// Do not handle page requests from bots and search engines
		if (!UnbEnvironment::IsBrowser()) return;

		// If we don't handle cross-domain requests, then this is a good time to exit
		if (!$this->crossDomainSupport && $_SERVER['HTTP_ORIGIN']) return;

		// Handle DNT header configuration
		if (UnbEnvironment::DoNotTrack())
		{
			$this->trackDntLevel = UnbConfiguration::Get('track.dnt.level', 0);
		}

		// Determine request type
		$isPageviewRequest = false;
		if ($_GET['imgns'])
			$isPageviewRequest = true;
		else if (!$_REQUEST[self::key_pageviewId])
			$isPageviewRequest = true;
		else
			$isPageviewRequest = false;
		
		// Check for admin optout cookie
		$optoutCookieName = UnbConfiguration::Get('cookie.name.optout', 'tomo_optout');
		$optoutToken = $_COOKIE[$optoutCookieName];
		$password = UnbConfiguration::Get('optout.password');
		$saltLength = 8;

		if (isset($optoutToken))
		{
			// Validate given optout token
			$salt = substr($optoutToken, 0, $saltLength);
			if ($optoutToken != $salt . md5($salt . $password))
			{
				// Invalid token, don't use it
				$optoutToken = null;
			}
		}

		if ($this->trackDntLevel < 4 && !isset($optoutToken))
		{
			// No (valid) opt-out cookie is set, continue tracking. Always send an appropriate
			// response though, even if it doesn't contain useful data.

			// Track the request
			try
			{
				// Try to get a lock on the client IP address
				$maxDelay = 5;   // Maximum time to wait in s
				$interval = 25;   // Retry interval in ms
				$retries = $maxDelay * (1000 / $interval);
				$haveLock = false;
				$t0 = microtime(true);
				while ($retries > 0)
				{
					if ($haveLock = $this->AcquireLock()) break;
					usleep($interval * 1000);
					$retries--;
				}
				if ($haveLock)
				{
					// Ready to go!
					$this->lockWaitTime = (microtime(true) - $t0) * 1000;
					UnbDebug::Trace('AcquireLock: success, waited ' . round($this->lockWaitTime, 2) . ' ms, lock id: ' . $this->lockId);
					$this->TrackRequest();
				}
				else
				{
					// Well, bad luck...
					UnbDebug::Trace('EXIT: Could not acquire lock on IP address in ' . $maxDelay . ' seconds');
					exit();
				}
			}
			catch (Exception $ex)
			{
				$errText = get_class($ex) . ': ' . $ex->getMessage() . ' (code ' . $ex->getCode() . ")\n" . UnbDebug::GetCallTrace($ex);
			}
			$this->ReleaseLock();

			// Log this request to the debug log
			$this->PostLog($errText, $isPageviewRequest);
		}

		if ($this->trackDntLevel < 2 && $this->newVisitorId)
		{
			// Only set the cookie if there's a need for it
			$days = intval(UnbConfiguration::Get('cookie.months', 12)) * 30.44;
			if ($this->trackDntLevel >= 1)
			{
				$days = 7;
			}
			$expire = time() + intval($days * 86400);
			
			setcookie(
				UnbConfiguration::Get('cookie.name.visitor', '_tid'),
				$this->EncodeVisitorIdForCookie($this->newVisitorId),
				$expire,
				UnbConfiguration::Get('cookie.path'));
		}

		if ($this->response[self::key_pageviewId])
		{
			// Send a short privacy description for the first request of a page (pageview ID is set)
			$privacy = '';
			if ($this->trackDntLevel >= 2)
				$privacy = 'No visitor cookie is set so we cannot recognise you next time around.';
			if ($this->trackDntLevel == 3)
				$privacy .= ' Also, activity timespan and in-page events will not be tracked.';
			if ($this->trackDntLevel >= 4)
				$privacy .= ' Also, page views will not be tracked and your data is not processed.';
			if ($privacy)
				$this->response['DNT_Info'] = 'Your privacy is respected: ' . $privacy;
		}

		if ($this->trackDntLevel >= 2 && $this->response[self::key_visitorId])
		{
			// Do not send a visitor ID cookie and also do not pass that information to the browser
			// so that it cannot store it selsewhere, like in DOM storage.
			$this->response[self::key_visitorId] = 0;
		}
		
		if ($this->trackDntLevel >= 3 && $this->response[self::key_pageviewId])
		{
			// Disable in-page event tracking for levels >= 3. This is done by not sending the
			// pageview ID for the current request. Without this ID, no more page events can be
			// sent.
			$this->response[self::key_pageviewId] = 0;
		}

		// Never cache tracker responses anywhere, each of them must arrive at the server
		header('Pragma: no-cache');
		header('Cache-Control: no-cache');
		header('Expires: 0');
		if ($this->crossDomainSupport)
			header('Access-Control-Allow-Origin: *');   // Enable for limited cross-domain support (not recommended)

		// Send the response to the browser
		if ($_GET['img'] || $_GET['imgns'])
		{
			// Send a small empty PNG image to the browser
			header('Content-Type: image/png');
			echo base64_decode('iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAYAAAAfFcSJAAAABmJLR0QA/wD/AP+gvaeTAAAAC0lEQVR42mNgAAIAAAUAAen63NgAAAAASUVORK5CYII=');
		}
		else if (count($this->response))
		{
			// Return information for further page action requests in JSON format
			header('Content-Type: application/json');
			echo json_encode($this->response);
		}
	}
	
	// Adds a message to the debug POST log.
	//
	private function PostLog($errText, $isPageviewRequest)
	{
		if (!$this->writePostLog) return;
		
		list($usec, $time) = explode(' ', microtime());
		$msg = date('Y-m-d H:i:s', $time) . '.' . substr($usec, 2, 6) . "\n";
		$msg .= "RequestTag: " . UnbDebug::GetRequestTag() . "\n";
		$msg .= "IP: " . UnbEnvironment::GetClientAddress() . "\n";
		$msg .= "Referrer: " . UnbEnvironment::GetReferrer() . "\n";
		$msg .= "User agent: " . $_SERVER['HTTP_USER_AGENT'] . "\n";
		$msg .= "GET:\n";
		foreach ($_GET as $name => $value)
		{
			$msg .= "  " . $name . " = " . str_replace("\n", "\n" . str_repeat(' ', strlen($name) + 5), $value) . "\n";
		}
		$msg .= "POST:\n";
		foreach ($_POST as $name => $value)
		{
			$msg .= "  " . $name . " = " . str_replace("\n", "\n" . str_repeat(' ', strlen($name) + 5), $value) . "\n";
		}
		$msg .= "HTTP:\n";
		foreach ($_SERVER as $name => $value)
		{
			if (substr($name, 0, 5) == 'HTTP_')
			{
				$name = strtolower(substr($name, 5));
				$name = str_replace('_', '-', $name);
				$msg .= "  " . $name . " = " . str_replace("\n", "\n" . str_repeat(' ', strlen($name) + 5), $value) . "\n";
			}
		}
		$msg .= "Visitor cookie: " . $this->GetVisitorIdFromCookie() . ' (' . $_COOKIE['_tid'] . ")\n";
		if ($errText)
		{
			$msg .= "ERROR: " . $errText . "\n";
		}
		$ms = round((microtime(true) - $this->startTime) * 1000, 2);
		$msg .= "Processing time : " . $ms . " ms" . ($ms > 100 ? ' *** LONG PROCESSING TIME ***' : '') . "\n";
		$msg .= "  Lock wait time: " . round($this->lockWaitTime, 2) . " ms\n";
		$msg .= "  DNS time      : " . round($this->dnsTime, 2) . " ms\n";

		if ($this->newVisitorId)
		{
			$msg .= "Set cookie: " . $this->newVisitorId . " (" . $this->EncodeVisitorIdForCookie($this->newVisitorId) . ")\n";
		}
		if (!$_GET['img'] && !$_GET['imgns'] && count($this->response))
		{
			$msg .= "JSON response:\n";
			foreach ($this->response as $key => $value)
			{
				$msg .= '  ' . $key . ' = ' . $value . "\n";
			}
		}
		
		$msg .= "\n";
		file_put_contents('post.' . date('Y-m-d', $time) . '.log', $msg, FILE_APPEND);
	}

	// Adds the page view to the database, with all posted data from the browser.
	//
	public function TrackRequest()
	{
		UnbDebug::Trace('ENTER TrackRequest');
		if ($_POST)
		{
			$rawData = $_POST;
		}
		else
		{
			$rawData = $_GET;
			#$rawData['_getRequest'] = true;   // DEBUG info only
		}

		if ($_GET['imgns'])
		{
			// Scripting is disabled on the browser, try to get some data from here.
			// Some values should have been passed with the GET parameters.
			$rawData['javascript'] = false;
			if (!$rawData[self::key_pageUrl])
			{
				$rawData[self::key_pageUrl] = $_SERVER['HTTP_REFERER'];
			}
		}
		else
		{
			$rawData['javascript'] = true;
		}

		// Check whether all data required to track the request is present
		if (!isset($rawData[self::key_pageviewId]) && !strlen($rawData[self::key_pageUrl]))
			throw new Exception('No pageview ID and page URL set in POST data');

		$now = time();

		$db = UnbDatabase::GetInstance();
		$db->StartTransaction();
		try
		{
			if (isset($rawData[self::key_pageviewId]))
			{
				// Pageview ID is set, this is an update or action request
				$pageviewId = intval($rawData[self::key_pageviewId]);
				$pageview = $this->FindPageviewById($pageviewId);
				if (isset($pageview))
				{
					// The pageview exists
					UnbDebug::Trace('  Branch: Pageview update request');
					$sessionId = intval($pageview['SessionId']);
					$session = $this->FindSessionById($sessionId);

					// Visitor ID cookie handling:
					$visitorId = $this->GetVisitorIdFromCookie();
					if (isset($visitorId) && $this->ExistsVisitorId($visitorId))
					{
						// The visitor ID cookie is set and valid
						if ($visitorId != $session['VisitorId'])
						{
							// Visitor ID from cookie does not match pageview's session visitor ID
							// Merge both visitors (keep reported, new ID from cookie)
							UnbDebug::Trace('    Merging two visitor entries');
							$this->MergeVisitors($visitorId, intval($session['VisitorId']));
							// Update session to get the new VisitorId value (required for session merging below)
							$session = $this->FindSessionById($sessionId);
						}
					}
					else
					{
						if (isset($visitorId))
							UnbDebug::Trace('    Noticed INVALID visitor cookie ' . $visitorId);
						else
							UnbDebug::Trace('    Noticed missing visitor cookie');
						// The visitor ID cookie is unset or invalid
						// Remember that cookies are not (correctly) supported
						if (!isset($visitorId) && $this->trackDntLevel < 2)
						{
							$rawData[self::key_cookies] = 0;
						}
						// Get visitor ID by pageview ID (via session)
						$visitorId = intval($session['VisitorId']);
						// Remember to send the visitor cookie to the client
						$this->newVisitorId = $visitorId;
					}
					// Now visitor ID of pageview and cookie (if set) match

					// Sessions handling:
					$currentSessions = $this->GetCurrentSessionsForVisitor($visitorId);
					// Logic and database constraints say that a session must exist here
					if (count($currentSessions) > 1)
					{
						UnbDebug::Trace('    Found multiple sessions, merging');
						// Multiple current sessions exist, merge them.
						// The array contains the current session, it will be skipped when we find it.
						for ($i = 0; $i < count($currentSessions); $i++)
						{
							if ($currentSessions[$i]['SessionId'] == $sessionId)
								// This is the session we're merging all others into, skip it.
								continue;

							$this->MergeSessions($session, $currentSessions[$i]);
							// Update the merged session data to keep it current for further merges
							$session = $this->FindSessionById($sessionId);
						}
					}

					$this->UpdateSession($session, $rawData, $now);

					// Pageviews handling:
					if (!$rawData[self::key_duplicatePageview] && $pageview['PageLoadTime'] !== '0')
					{
						// Find duplicates of the current pageview using the pageview data already
						// loaded by the reported pageview ID. Don't merge anything with a virtual
						// pageview reported through the JavaScript API.
						$pageviews = $this->FindDuplicatePageviews($pageview);
						if (count($pageviews))
						{
							UnbDebug::Trace('    Found multiple pageviews, merging');
							// Duplicate pageviews exist, merge them (keep the reported ID)
							// The array does not contain the current pageview, so merge them all.
							for ($i = 0; $i < count($pageviews); $i++)
							{
								$this->MergePageviews($pageview, $pageviews[$i]);
								// Update the merged pageview data to keep it current for further merges
								$pageview = $this->FindPageviewById($pageviewId);
							}
						}
					}

					UnbDebug::Trace('    Updating pageview');
					$this->UpdatePageview($pageview, $rawData, $now);

					// Actions handling:
					if ($rawData[self::key_outlinkUrl])
					{
						// This event reports an outlink action
						$type = self::action_Outlink;
						$url = $rawData[self::key_outlinkUrl];
						$data = null;
						$actions = $this->FindDuplicateActions($now, $pageviewId, $type, $url, $data);
						if (count($actions) == 0)
						{
							// No duplicate actions exist yet, go ahead and create it
							$this->CreateAction($now, $pageviewId, $type, $url, $data);
						}
					}
					if ($rawData[self::key_downloadUrl])
					{
						// This event reports a download action
						$type = self::action_Download;
						$url = $rawData[self::key_downloadUrl];
						$data = null;
						$actions = $this->FindDuplicateActions($now, $pageviewId, $type, $url, $data);
						if (count($actions) == 0)
						{
							// No duplicate actions exist yet, go ahead and create it
							$this->CreateAction($now, $pageviewId, $type, $url, $data);
						}
					}
					if ($rawData[self::key_imgZoomUrl])
					{
						// This event reports an image zoom action
						$type = self::action_ImgZoom;
						$url = $rawData[self::key_imgZoomUrl];
						$data = null;
						$actions = $this->FindDuplicateActions($now, $pageviewId, $type, $url, $data);
						if (count($actions) == 0)
						{
							// No duplicate actions exist yet, go ahead and create it
							$this->CreateAction($now, $pageviewId, $type, $url, $data);
						}
					}
					if ($rawData[self::key_sectionExpand])
					{
						// This event reports an image zoom action
						$type = self::action_SectionExpand;
						$url = null;
						$data = $rawData[self::key_sectionExpand];
						$actions = $this->FindDuplicateActions($now, $pageviewId, $type, $url, $data);
						if (count($actions) == 0)
						{
							// No duplicate actions exist yet, go ahead and create it
							$this->CreateAction($now, $pageviewId, $type, $url, $data);
						}
					}
					if ($rawData[self::key_siteSearch])
					{
						// This event reports an image zoom action
						$type = self::action_SiteSearch;
						$url = null;
						$data = $rawData[self::key_siteSearch];
						$actions = $this->FindDuplicateActions($now, $pageviewId, $type, $url, $data);
						if (count($actions) == 0)
						{
							// No duplicate actions exist yet, go ahead and create it
							$this->CreateAction($now, $pageviewId, $type, $url, $data);
						}
					}

					$confActions = self::GetConfActions();
					foreach ($confActions as $name => $type)
					{
						if (isset($rawData['ac_' . $name . '_u']) || isset($rawData['ac_' . $name . '_d']))
						{
							$url = $rawData['ac_' . $name . '_u'];
							$data = $rawData['ac_' . $name . '_d'];
							if (!strlen($data))
								$data = null;
							$actions = $this->FindDuplicateActions($now, $pageviewId, $type, $url, $data);
							if (count($actions) == 0)
							{
								// No duplicate actions exist yet, go ahead and create it
								$this->CreateAction($now, $pageviewId, $type, $url, $data);
							}
						}
					}

					if ($rawData[self::key_duplicatePageview] && $rawData['javascript'])
					{
						// Duplicating a pageview is only possible with JavaScript
						UnbDebug::Trace('  Pageview duplicate request');
						
						// Process rawData for second pageview record
						$newRawData = array();
						foreach ($rawData as $key => $value)
						{
							if (preg_match('/^n_(.+)$/', $key, $m))
							{
								$newRawData[$m[1]] = $value;
							}
						}
						
						$pageview = $this->DuplicatePageview($pageview, $newRawData, $now);
						$this->response[self::key_pageviewId] = intval($pageview['PageviewId']);
					}
				}
				else
				{
					UnbDebug::Trace('  Branch: INVALID pageview update request');
					// The pageview does not exist, do nothing more
					// Abort request here, actually, throw an exception or something. All instance data is invalid.
					throw new Exception('Invalid pageview update request: Pageview ID is invalid');
				}
			}
			else
			{
				UnbDebug::Trace('  Branch: Page enter request, js=' . ($rawData['javascript'] ? 1 : 0));
				// Pageview ID is not set, this is a page enter request
				// Visitor ID cookie handling:
				$visitorId = $this->GetVisitorIdFromCookie();
				if (isset($visitorId) && $this->ExistsVisitorId($visitorId))
				{
					// The visitor ID cookie is set and valid
					if (!$rawData['javascript'])
					{
						// JavaScript data is not available, so the cookies test has not run. But
						// we see a valid cookie here, so cookies support can be assumed.
						$rawData[self::key_cookies] = 1;
					}
				}
				else
				{
					// The visitor ID cookie is unset or invalid
					$visitorId = null;
				}
				// Now visitor ID is either valid or unset
				
				if (!$visitorId)
				{
					// Visitor is yet unknown, try to recognise by similar sessions
					$sessionData = $this->PrepareSessionData($rawData, $now, 0, false);
					$session = $this->FindSimilarSession($sessionData);
					if ($session)
					{
						UnbDebug::Trace('    Visitor identified by similar session data');
						$visitorId = intval($session['VisitorId']);
						// Remember to send the visitor cookie to the client
						$this->newVisitorId = $visitorId;
					}
					else if (intval($rawData[self::key_origPageviewId]) > 0)
					{
						// The request contains an original pageview ID, so it comes from the very
						// same window that an earlier request came from. We can use the other
						// pageview for session identification.
						$origPVId = intval($rawData[self::key_origPageviewId]);
						$pageview = $this->FindPageviewById($origPVId);
						if (isset($pageview) &&
							$now < UnbDatabase::ConvertDateTime($pageview['PageviewTime']) + 86400)
						{
							// The specified pageview exists and is not older than 1 day
							$sessionId = intval($pageview['SessionId']);
							$session = $this->FindSessionById($sessionId);
							$visitorId = intval($session['VisitorId']);
							// Remember to send the visitor cookie to the client
							$this->newVisitorId = $visitorId;
						}
					}
				}
				
				if (!$visitorId)
				{
					UnbDebug::Trace('    Visitor cannot be identified -> create visitor, session and pageview');
					// Visitor is not set and none can be found by similar sessions.
					// Create new visitor, session and pageview
					$visitorId = $this->CreateVisitor();
					$this->response[self::key_visitorId] = $visitorId;
					$this->newVisitorId = $visitorId;
					$this->response[self::key_newVisitor] = true;

					$session = $this->CreateSession($rawData, $now, $visitorId, true);
					$this->response[self::key_newSession] = true;

					$pageview = $this->CreatePageview($rawData, $now, intval($session['SessionId']), true);
					$this->response[self::key_pageviewId] = intval($pageview['PageviewId']);
				}
				else
				{
					// Visitor is known.
					$this->response[self::key_visitorId] = $visitorId;
					$this->response[self::key_newVisitor] = false;

					// Find current sessions
					$currentSessions = $this->GetCurrentSessionsForVisitor($visitorId);
					if (!$currentSessions)
					{
						UnbDebug::Trace('    No active sessions -> Create session and pageview');
						// No current sessions for this visitor.
						// This is a new visit of a known visitor.
						// Create new session and pageview
						$session = $this->CreateSession($rawData, $now, $visitorId, false);
						$this->response[self::key_newSession] = true;

						$pageview = $this->CreatePageview($rawData, $now, intval($session['SessionId']), true);
						$this->response[self::key_pageviewId] = intval($pageview['PageviewId']);
					}
					else
					{
						UnbDebug::Trace('    ' . count($currentSessions) . ' active session(s) found');
						// Current session(s) exist(s).
						$session = $currentSessions[0];
						$sessionId = intval($session['SessionId']);
						if (count($currentSessions) > 1)
						{
							// Multiple current sessions exist, merge them.
							// The array contains all current sessions, newest first.
							// Skip the first result and merge all others into that.
							for ($i = 1; $i < count($currentSessions); $i++)
							{
								$this->MergeSessions($session, $currentSessions[$i]);
								// Update the merged session data to keep it current for further merges
								$session = $this->FindSessionById($sessionId);
							}
						}

						$this->UpdateSession($session, $rawData, $now);
						$this->response[self::key_newSession] = false;
					
						// Find duplicates of the current pageview data
						$pageviewData = $this->PreparePageviewData($rawData, $now, $sessionId, false);
						$pageviews = $this->FindDuplicatePageviews($pageviewData);
						$pageview = $pageviews[0];
						if (count($pageviews))
						{
							UnbDebug::Trace('    ' . count($pageviews) . ' duplicate pageview(s) found');
						}
						if (count($pageviews) > 1)
						{
							// Multiple duplicate pageviews exist, first merge them.
							// The array contains anything (the new pageview is not yet stored), so merge them all.
							$pageviewId = intval($pageview['PageviewId']);
							for ($i = 1; $i < count($pageviews); $i++)
							{
								$this->MergePageviews($pageview, $pageviews[$i]);
								// Update the merged pageview data to keep it current for further merges
								$pageview = $this->FindPageviewById($pageviewId);
							}
						}

						if (count($pageviews))
						{
							UnbDebug::Trace('    Existing pageview ' . $pageview['PageviewId'] . ' updated');
							// Any duplicate pageview already exists (multiple would have been merged into one).
							// Update that pageview.
							$this->UpdatePageview($pageview, $rawData, $now);
							// Send the already stored activity timespan to the client to continue
							// counting.
							$this->response[self::key_activityTimespan] = intval($pageview['ActivityTimespan']) * 1000;
						}
						else
						{
							UnbDebug::Trace('    Pageview created');
							// No duplicate pageview exists yet, create a new one.
							$pageview = $this->CreatePageview($rawData, $now, $sessionId, false);
						}
						$this->response[self::key_pageviewId] = intval($pageview['PageviewId']);
					}
				}
			}

			$db->Commit();
		}
		catch (Exception $ex)
		{
			$db->RollBack();
			throw $ex;
		}
	}


	// ====================================  COOKIE FUNCTIONS  ====================================
	

	// Gets the visitor ID from the cookie, if present.
	//
	// Returns (int) decoded visitor ID, or {{null}}
	//
	private function GetVisitorIdFromCookie()
	{
		$vCookieName = UnbConfiguration::Get('cookie.name.visitor', '_tid');
		$vCookie = $_COOKIE[$vCookieName];

		if (!strlen($vCookie)) return null;
		
		//return intval(hexdec($vCookie));
		#return intval($vCookie);   // DEBUG
		$data = @unpack('N', base64_decode($vCookie));
		$visitorId = $data[1];
		if ($visitorId <= 0) return null;
		return $visitorId;
	}

	// Gets the visitor ID encoded for use in a cookie.
	//
	// visitorId = (int) Visitor ID to encode
	//
	// Returns (string) encoded visitor ID
	//
	private function EncodeVisitorIdForCookie($visitorId)
	{
		if (!is_int($visitorId))
			throw new InvalidArgumentException('Invalid argument type: $visitorId');
		if ($visitorId <= 0)
			throw new InvalidArgumentException('Argument out of range: $visitorId');

		//return str_pad(dechex($visitorId), 8, '0', STR_PAD_LEFT);
		#return '' . $visitorId;   // DEBUG
		return rtrim(base64_encode(pack('N', $visitorId)), '=');
	}


	// ===============================  DATA PREPARATION FUNCTIONS  ===============================
	

	// Reads all location-related values from the posted data array.
	//
	// Returns an array of values by their database column for the session table.
	//
	// Some of the values are extracted from the posted data, others are recognised from the
	// execution environment like the IP address and the geo IP lookup.
	//
	private function GetLocationSessionValues($rawData)
	{
		if (isset($this->locationValuesCache)) return $this->locationValuesCache;
		
		$ipAddress = UnbEnvironment::GetClientAddress();
		$ipVersion = strpos($ipAddress, '.') !== false ? 4 : 6;
		
		$locData = GeoIP::LookupIP($ipAddress);
		$countryCode = $locData['cc'];
		$region = $locData['region'];
		$city = strlen($locData['city']) ? utf8_encode($locData['city']) : null;
		$lat = $locData['lat'];
		$lng = $locData['lng'];

		$provider = null;
		$t0 = microtime(true);
		$hostName = gethostbyaddr($ipAddress);
		$this->dnsTime = (microtime(true) - $t0) * 1000;
		$providerTld = null;
		if ($hostName != $ipAddress)
		{
			// IP address could be resolved, now extract the best-guess of a provider name
			$parts = explode('.', $hostName);
			$provider = array_pop($parts);
			$providerTld = $provider;
			$provider = PublicSuffix::GetPublicSuffix($hostName);
			if ($provider === false)
			{
				// An error occured in the public suffix library, use our old simple method
				$provider = $providerTld;
				$next = array_pop($parts);
				if ($next) $provider = $next . '.' . $provider;
				if (in_array($next, array('co', 'com', 'gov', 'net', 'org')) ||
					$providerTld == 'ch' && strlen($next) == 2 ||
					$providerTld == 'uk' && in_array($next, array('ac', 'ltd', 'me', 'nhs', 'plc', 'sch')))
				{
					$next = array_pop($parts);
					if ($next) $provider = $next . '.' . $provider;
				}
			}
		}
		
		// Don't record a proxy network's public endpoint
		if ($provider == 'opera-mini.net')
		{
			$countryCode = null;
			$region = null;
			$city = null;
			$lat = null;
			$lng = null;
		}

		// Validate country data
		$ccGood = self::ValidateCountry(
			$countryCode,
			$region,
			$city,
			$lat,
			$lng,
			$rawData[self::key_utcOffset],
			$providerTld,
			UnbEnvironment::GetClientLanguages(),
			$rawData[self::key_referrer]);

		$values = array(
			'IPAddress' => $ipAddress,
			'IPVersion' => $ipVersion,
			'CountryCode' => $countryCode,
			);
		if ($ccGood)
		{
			if (strlen($region))
			{
				$values['RegionText'] = Tomo::GetTextId($region);
			}
			if (strlen($city))
			{
				$values['CityText'] = Tomo::GetTextId($city);
			}
			if (isset($lat) && isset($lng))
			{
				$values['Lat'] = $lat * 10000;
				$values['Lng'] = $lng * 10000;
			}
		}
		else
		{
			// Country has been found invalid and changed into something else.
			// We don't have valid region data anymore, so delete it.
			$values['RegionText'] = null;
			$values['CityText'] = null;
			$values['Lat'] = null;
			$values['Lng'] = null;
		}
		if (isset($provider))
		{
			$values['ProviderText'] = Tomo::GetTextId($provider);
		}

		$this->locationValuesCache = $values;
		return $values;
	}

	// Reads all referrer-related values from the posted data array.
	//
	// Returns an array of values by their database column for the session table.
	//
	private function GetReferrerSessionValues($rawData)
	{
		if (isset($this->referrerValuesCache)) return $this->referrerValuesCache;

		$values = array();
		$oref = $this->GetOriginalRefFromPageUrl($rawData[self::key_pageUrl]);
		if (isset($oref))
		{
			// Old-site redirection. This parameter contains the original referrer if it is still
			// available. We notice this referrer instead of the URL of our own old site which is
			// redirecting here. The old site's domain is set as campaign name.
			$refUrl = parse_url($rawData[self::key_referrer]);
			$values['CampaignText'] = Tomo::GetTextId($refUrl['host']);
			UnbDebug::Trace('      Old-site redirection (Campaign): ' . $refUrl['host']);
			// Replace the referrer with the URL parameter value
			$rawData[self::key_referrer] = $oref;
		}

		if (strlen($rawData[self::key_referrer]))
		{
			// Split referrer URL in SSL, hostname/port, and path/query/hash
			$refUrl = parse_url($rawData[self::key_referrer]);
			$refSsl = $refUrl['scheme'] == 'https';
			$refHost = $refUrl['host'] . ($refUrl['port'] ? ':' . $refUrl['port'] : '');
			$refPath = $refUrl['path'];
			if (isset($refUrl['query']))
				$refPath .= '?' . $refUrl['query'];
			if (isset($refUrl['fragment']))
				$refPath .= '#' . $refUrl['fragment'];

			$searchTerms = UnbEnvironment::GetSearchTerms($rawData[self::key_referrer]);
			if ($searchTerms)
			{
				$values['SearchKeywordsText'] = Tomo::GetTextId($searchTerms);
			}
			if (UnbEnvironment::GetSearchEngine())   // Must be called after GetSearchTerms
			{
				$values['SearchEngineText'] = Tomo::GetTextId(UnbEnvironment::GetSearchEngine());

				// Strip the search engine referrer URL down to their domain only to save a lot of space
				$refPath = '';
			}
			if (UnbEnvironment::GetSearchImageUrl())   // Must be called after GetSearchTerms
			{
				$values['SearchImageUrlText'] = Tomo::GetTextId(UnbEnvironment::GetSearchImageUrl());
			}
			if (UnbEnvironment::GetSearchPageNumber())   // Must be called after GetSearchTerms
			{
				$values['SearchResultPage'] = UnbEnvironment::GetSearchPageNumber();
			}

			$values['ReferrerSsl'] = $refSsl ? 1 : 0;
			$values['ReferrerHostText'] = Tomo::GetTextId($refHost);
			$values['ReferrerUrlPathText'] = Tomo::GetTextId($refPath);
		}
		if (strlen($rawData[self::key_campaign]))
		{
			$values['CampaignText'] = Tomo::GetTextId($rawData[self::key_campaign]);
			UnbDebug::Trace('      Campaign: ' . $rawData[self::key_campaign]);
		}
		else
		{
			$campaign = $this->GetCampaignFromPageUrl($rawData[self::key_pageUrl]);
			if ($campaign)
			{
				$values['CampaignText'] = Tomo::GetTextId($campaign);
				UnbDebug::Trace('      Campaign: ' . $campaign);
			}
		}

		$this->referrerValuesCache = $values;
		return $values;
	}
	
	// Analyses the specified page URL and finds a campaign name if present.
	//
	// url = (string) URL of the page being displayed.
	//
	// Returns (string) The campaign name, or null.
	//
	private function GetCampaignFromPageUrl($url)
	{
		$pageUrl = parse_url($url);
		parse_str(ltrim($pageUrl['query'], '?'), $vars);
		return $vars['ref'];
	}

	// Analyses the specified page URL and finds the original referrer if present.
	//
	// url = (string) URL of the page being displayed.
	//
	// Returns (string) The original referrer, or null.
	//
	private function GetOriginalRefFromPageUrl($url)
	{
		$pageUrl = parse_url($url);
		parse_str(ltrim($pageUrl['query'], '?'), $vars);
		return $vars['oref'];
	}

	// Reads all browser-related values from the posted data array.
	//
	// Returns an array of values by their database column for the session table.
	//
	// Some of the values are extracted from the posted data, others are recognised from the
	// execution environment like HTTP request headers.
	//
	private function GetBrowserSessionValues($rawData)
	{
		if (isset($this->browserValuesCache)) return $this->browserValuesCache;

		$language = '';
		$langs = UnbEnvironment::GetClientLanguages();
		if (count($langs))
		{
			switch (UnbConfiguration::Get('track.language.level', 2))
			{
				case 1:
					$language = substr($langs[0], 0, 2);
					break;
				case 2:
				default:
					$language = $langs[0];
					break;
				case 3:
					$language = join(',', $langs);
					break;
			}
		}

		$platform = UnbEnvironment::GetClientPlatform();
		if (strpos($platform, '@') !== false)
		{
			list($browser, $os) = explode('@', $platform);
		}
		else
		{
			$browser = $platform;
			$os = null;
		}

		$values = array(
			'BrowserText' => self::GetTextId($browser),
			'OperatingSystemText' => self::GetTextId($os),
			'Mobile' => UnbEnvironment::IsMobile() ? 1 : 0,
			'Language' => $language,
			'JavaScript' => $rawData['javascript'] ? 1 : 0,
			);
		$dnt = UnbEnvironment::DoNotTrack();
		if (isset($dnt))
		{
			$values['DNT'] = $dnt ? 1 : 0;
		}

		if (strlen($rawData[self::key_utcOffset]))
		{
			$values['UtcOffset'] = intval($rawData[self::key_utcOffset]);
		}
		if (strlen($rawData[self::key_cookies]))
		{
			$values['Cookies'] = $rawData[self::key_cookies] ? 1 : 0;
		}
		if (strlen($rawData[self::key_flash]))
		{
			$values['Flash'] = $rawData[self::key_flash] ? 1 : 0;
		}
		if (strlen($rawData[self::key_java]))
		{
			$values['Java'] = $rawData[self::key_java] ? 1 : 0;
		}
		if (strlen($rawData[self::key_quicktime]))
		{
			$values['Quicktime'] = $rawData[self::key_quicktime] ? 1 : 0;
		}
		if (strlen($rawData[self::key_silverlight]))
		{
			$values['Silverlight'] = $rawData[self::key_silverlight] ? 1 : 0;
		}
		if (strlen($rawData[self::key_windowsmedia]))
		{
			$values['WindowsMedia'] = $rawData[self::key_windowsmedia] ? 1 : 0;
		}
		if (strlen($rawData[self::key_adblock]))
		{
			$values['AdBlock'] = $rawData[self::key_adblock] ? 1 : 0;
		}
		if ($rawData[self::key_screenWidth] > 0)
		{
			$values['ScreenWidth'] = intval($rawData[self::key_screenWidth]);
		}
		if ($rawData[self::key_screenHeight] > 0)
		{
			$values['ScreenHeight'] = intval($rawData[self::key_screenHeight]);
		}
		if ($rawData[self::key_screenAvailWidth] > 0)
		{
			$values['ScreenAvailWidth'] = intval($rawData[self::key_screenAvailWidth]);
		}
		if ($rawData[self::key_screenAvailHeight] > 0)
		{
			$values['ScreenAvailHeight'] = intval($rawData[self::key_screenAvailHeight]);
		}
		if ($rawData[self::key_screenColors] > 0)
		{
			$values['ScreenColors'] = intval($rawData[self::key_screenColors]);
		}
		if ($rawData[self::key_viewportWidth] > 0)
		{
			$values['ViewportWidth'] = intval($rawData[self::key_viewportWidth]);
		}
		if ($rawData[self::key_viewportHeight] > 0)
		{
			$values['ViewportHeight'] = intval($rawData[self::key_viewportHeight]);
		}

		$this->browserValuesCache = $values;
		return $values;
	}
	
	// Comparison helper method to find which information to keep when merging two sessions.
	//
	private function MergeSessionHelper($session1, $session2, $field, $minmax, $preferJs)
	{
		$v1 = $session1[$field];
		$v2 = $session2[$field];
		// Take one if both are equal
		if ($v1 === $v2)
			return $v1;
		// Take the set one if only one is set
		if (isset($v1) && !isset($v2))
			return $v1;
		if (!isset($v1) && isset($v2))
			return $v2;
		// Now both are set, is JavaScript preferred?
		if ($preferJs)
		{
			if ($session1['JavaScript'] && !$session2['JavaScript'])
				return $v1;
			if (!$session1['JavaScript'] && $session2['JavaScript'])
				return $v2;
			// Now neither or both sessions have JavaScript
		}
		if ($minmax == 0)
			return $min($v1, $v2);
		if ($minmax == 1)
			return $max($v1, $v2);
		// Here be dragons...
		return $v1;
	}

	// Converts the raw reported data to field values to be used in the database.
	//
	// This is used by CreatePageview and can also be used to create similar data to pass to
	// FindDuplicatePageviews.
	//
	// Returns an array of the prepared pageview data (without PageviewId).
	//
	private function PreparePageviewData($rawData, $now, $sessionId, $entryPage)
	{
		if (!strlen($rawData[self::key_pageUrl]))
			throw new Exception('Page URL not set in POST data');
		
		$values = array(
			'SessionId' => $sessionId,
			'PageviewTime' => UnbDatabase::ConvertDateTime($now),
			'PageviewUpdateTime' => UnbDatabase::ConvertDateTime($now),
			'EntryPage' => $entryPage ? 1 : 0,
			'JavaScriptData' => $rawData['javascript'] ? 1 : 0,
			);

		// Parse page URL into its elements
		$pageUrl = parse_url($rawData[self::key_pageUrl]);
		// TODO: pageUrl: Only keep certain parameters in the query string and remove the others
		//       (Submit whitelist for that with the tracking request)

		$pattern = UnbConfiguration::Get('track.url.remove-path-pattern');
		if ($pattern)
			$pageUrl['path'] = preg_replace($pattern, '', $pageUrl['path']);

		if (strlen($rawData[self::key_pageTitle]))
		{
			$values['PageTitleText'] = Tomo::GetTextId($rawData[self::key_pageTitle]);
		}
		$langPath = UnbConfiguration::Get('track.language.first-path', false);
		if ($langPath && preg_match('~^/([a-z]{2}([-_][a-z]{2})?)(/|$)~i', $pageUrl['path'], $m))
		{
			$values['PageLanguage'] = $m[1];
			$pageUrl['path'] = substr($pageUrl['path'], strlen($m[1]) + 1);
		}
		if (strlen($rawData[self::key_pageLanguage]))
		{
			// Let this setting override our path detection because a website translation service
			// may be used.
			$values['PageLanguage'] = $rawData[self::key_pageLanguage];
		}

		if (isset($rawData[self::key_activityTimespan]))
		{
			$activityTimespan = intval($rawData[self::key_activityTimespan]);
			// Set the rounding threshold to 0.8, i.e. count a full second after 800 ms of it.
			$activityTimespan = intval(($activityTimespan + 200) / 1000);
			$pageviewTime = UnbDatabase::ConvertDateTime($pageview['PageviewTime']);
			if ($activityTimespan > $now - $pageviewTime + 2)
			{
				UnbDebug::Trace('  Reported activity timespan is larger than absolute pageview lifetime. Limit value.');
				$activityTimespan = $now - $pageviewTime;
			}
			$values['ActivityTimespan'] = $activityTimespan;
		}
		if ($rawData[self::key_pageLoadTime] != 0)
		{
			$values['PageLoadTime'] = intval($rawData[self::key_pageLoadTime]);
		}
		if (isset($rawData[self::key_prerender]))
		{
			$values['Prerender'] = $rawData[self::key_prerender] ? 1 : 0;
		}

		if (!strlen($pageUrl['path']))
			$pageUrl['path'] = '/';

		$values['DomainText'] = Tomo::GetTextId($pageUrl['host'] . ($pageUrl['port'] ? ':' . $pageUrl['port'] : ''));
		$values['UrlPathText'] = Tomo::GetTextId($pageUrl['path']);
		$values['UrlQueryText'] = Tomo::GetTextId($pageUrl['query']);
		$values['UrlHashText'] = Tomo::GetTextId($pageUrl['fragment']);
		$values['Ssl'] = $pageUrl['scheme'] == 'https' ? 1 : 0;
		
		if (strlen($rawData[self::key_campaign]) && $rawData['javascript'])
		{
			$values['CampaignRedirect'] = 1;
		}
		else
		{
			// Detect campaign parameter to be sure (for no-Js requests), always set this field (NOT NULL)!
			$values['CampaignRedirect'] = $this->GetCampaignFromPageUrl($rawData[self::key_pageUrl]) != null ? 1 : 0;
		}

		if (intval($rawData[self::key_origPageviewId]) > 0)
		{
			$values['OriginalPageviewId'] = intval($rawData[self::key_origPageviewId]);
		}

		return $values;
	}


	// ====================================  COUNTRY FUNCTIONS  ===================================
	

	// Validates and corrects the country and location result.
	//
	// inout countryCode = (string) Country code to validate; may be set to another value
	// out region = (string) Region name; may be set to null if the country code was changed
	// out city = (string) City name; may be set to null if the country code was changed
	// out lat = (float) Latitude; may be set to null if the country code was changed
	// inout lng = (float) Longitude; may be set to null if the country code was changed
	// utcOffset = (int string) UTC offset in minutes as reported by the browser
	// providerTld = (string) Hostname top-level domain, if available by DNS
	// langs = (array) Client languages array
	// referrer = (string) HTTP referrer URL
	//
	// Returns (bool) true if the country information is good, false if it was set to another
	// value. If this function returns false, the values of $region, $city, $lat and $lng are
	// considered invalid and must be deleted from the database.
	//
	private static function ValidateCountry(&$countryCode, &$region, &$city, &$lat, &$lng, $utcOffset, $providerTld, $langs, $referrer)
	{
		if (!isset($countryCode) ||
			isset($lng) && isset($utcOffset))
		{
			$utcOffset = intval($utcOffset);
			$diff = $lng / 15 - $utcOffset / 60;
			// TODO: Be less restrictive with time zone offsets for certain large
			//       countries if no region or city are available from GeoIP.
			if (!isset($countryCode) ||
				$diff < -2 || $diff > 4)
			{
				if (!isset($countryCode))
				{
					UnbDebug::Trace('      Country not detected by GeoIP database. Looking for alternatives...');
				}
				else
				{
					UnbDebug::Trace('      Country "' . $countryCode . '" longitude ' . $lng . ' does not match time zone ' . ($utcOffset / 60) . '. Looking for alternatives...');
				}
				// The difference between time zone and longitude is too great to believe.
				// Our IP-location database isn't perfect. Try get the country elsewhere.
				$hostnameCC = null;
				if (strlen($providerTld) == 2)
				{
					// The reverse DNS record has a ccTLD. Use that.
					$hostnameCC = strtoupper($providerTld);
					if ($hostnameCC == 'UK') $hostnameCC = 'GB';   // UK special case
					if (strtoupper($providerTld) != $countryCode)
					{
						UnbDebug::Trace('      Different country code from rdns name: ' . $hostnameCC);
					}
					else
					{
						UnbDebug::Trace('      rdns name gives the same country, so keep it.');
					}
				}
				$localeCC = null;
				if (count($langs) >= 1 && strlen($langs[0]) == 5)
				{
					// First language has a country locale part. Use that.
					$localeCC = substr($langs[0], 3);
					if (substr($langs[0], 3) != $countryCode)
					{
						UnbDebug::Trace('      Different country code from first language locale: ' . $localeCC);
					}
					else
					{
						UnbDebug::Trace('      First language locale gives the same country, so keep it.');
					}
				}
				else if (count($langs) >= 2 &&
					strlen($langs[0]) == 2 &&
					strlen($langs[1]) == 5 &&
					$langs[0] == substr($langs[1], 0, 2))
				{
					// First language has no country part but is the same language as the second
					// element which has a country part. Use that.
					$localeCC = substr($langs[1], 3);
					if (substr($langs[1], 3) != $countryCode)
					{
						UnbDebug::Trace('      Different country code from second language locale: ' . $localeCC);
					}
					else
					{
						UnbDebug::Trace('      Second language locale gives the same country, so keep it.');
					}
				}
				$googleCC = null;
				if (preg_match('_^https?://www\.google(?:\.(?:com?|net|org))?\.([a-z]{2})/(?:url|search|m|imgres)_', $referrer, $m))
				{
					// Referrer is a country-specific Google search. Use that.
					$googleCC = strtoupper($m[1]);
					if ($googleCC == 'UK') $googleCC = 'GB';   // UK special case
					if (strtoupper($m[1]) != $countryCode)
					{
						UnbDebug::Trace('      Different country code from country-specific Google search: ' . $googleCC);
					}
					else
					{
						UnbDebug::Trace('      Country-specific Google search gives the same country, so keep it.');
					}
				}

				// Determine the best found country code.
				$newCC = null;
				if (isset($hostnameCC))
				{
					// The hostname TLD has the same meaning as the IP's GeoIP entry, so use that
					// whenever we can get it.
					$newCC = $hostnameCC;
				}
				else if (isset($localeCC) && isset($googleCC))
				{
					if ($localeCC == $googleCC)
					{
						// Locale matches Google, pretty confident now...
						$newCC = $localeCC;
					}
					else
					{
						// Accept the more specific country of both, i.e. the smaller one.
						$ccs = array($localeCC, $googleCC);
						usort($ccs, array(__CLASS__, 'SpecificCountryCodeComparer'));
						$newCC = $ccs[0];
					}
				}
				else if (isset($localeCC))
				{
					$newCC = $localeCC;
				}
				else if (isset($googleCC))
				{
					$newCC = $googleCC;
				}

				// Verify the new country with the visitor's timezone, if available.
				if (isset($newCC) &&
					isset($utcOffset))
				{
					// Find the new country's longitude from GeoIP
					// TODO: This feature is not yet available by the GeoIP library.
					#$newLng = 0;
					#$diff = $newLng / 15 - $utcOffset / 60;
					#if ($diff < -2 || $diff > 4)
					#{
					#	// The new country code is also too far off the time zone. Dismiss it.
					#	$newCC = null;
					#}
					// TODO: Be less restrictive with time zone offsets for certain large
					//       countries if no region or city are available from GeoIP.
				}

				// Accept whatever new country code we have found and clear the other location
				// fields if the country has changed in the process.
				if (isset($newCC) && $newCC != $countryCode)
				{
					$countryCode = $newCC;
					$region = null;
					$city = null;
					$lat = null;
					$lng = null;
					return false;
				}
			}
		}
		return true;
	}

	// Gets a country specificity index by the country code. This value is used to sort countries
	// by their specificity, i.e. find the most specific country among multiple countries that
	// share the same language or even the same locale as browser presets.
	//
	// Basically, the number of inhabitants in a country is a good indicator for its specificity.
	// But it must be ensured that some country relations remain intact disregarding the number of
	// inhabitants. That menas, a country that is more often used a locale identifier, also in
	// other countries, must always be less specific.
	//
	// cc = (string) Country code to get the specificity for. Must be two-letter upper-case.
	//
	// Returns an integer value. Values only have relative meaning. Lower values are more specific.
	//
	private static function GetCountryCodeSpecificity($cc)
	{
		// Return approx. number of country inhabitants in thousands, see Wikipedia.
		switch ($cc)
		{
			case 'AR': return 40500;
			case 'AT': return 8400;   // AT must be more specific than DE
			case 'AU': return 22300;   // AU must be more specific than GB
			case 'BE': return 10800;
			case 'BR': return 195000;   // BR must be more specific than PT
			case 'CA': return 34300;   // CA must be more specific than US
			case 'CH': return 7900;   // CH must be more specific than DE
			case 'CN': return 1339700;
			case 'DE': return 81700;
			case 'ES': return 47100;
			case 'FR': return 65400;
			case 'GB': return 61800;
			case 'HK': return 7100;   // HK must be more specific than CN
			case 'MX': return 112300;
			case 'NL': return 16700;
			case 'NZ': return 4400;
			case 'PT': return 10600;
			case 'TW': return 9000;   // TW must be more specific than CN
			case 'US': return 311500;
			// Anything else is probably more specific, but we may want to extend this list a bit further.
			default: 0;
		}
	}

	// Comparer function to sort countries by their specificity.
	//
	private static function SpecificCountryCodeComparer($a, $b)
	{
		$sa = self::GetCountryCodeSpecificity($a);
		$sb = self::GetCountryCodeSpecificity($b);
		return $sa - $sb;
	}


	// ===================================  DATABASE FUNCTIONS  ===================================
	
	
	// -------------------------------  Database: Visitor functions  ------------------------------
	
	
	// Gets a value indicating whether the specified visitor ID exists.
	//
	private function ExistsVisitorId($visitorId)
	{
		if (!is_int($visitorId))
			throw new InvalidArgumentException('Invalid argument type: $visitorId');
		if ($visitorId < 0)
			throw new InvalidArgumentException('Argument out of range: $visitorId');

		$db = UnbDatabase::GetInstance();
		$rec = $db->QueryRecord('SELECT "VisitorId" FROM "#visitor" WHERE "VisitorId" = ?', $visitorId);
		return $rec ? true : false;
	}
	
	// Creates a new visitor in the database.
	//
	// Finds and occupies a new visitor ID number.
	//
	// Returns (int) new assigned visitor ID.
	//
	private function CreateVisitor()
	{
		$db = UnbDatabase::GetInstance();
		$db->StartTransaction();
		try
		{
			// Loop to allow to repeat the ID selection process if it should fail in the very last
			// moment, ie. when someone else had the same idea and created the selected ID a
			// moment before. Normally, this loop is executed just once then left. To avoid an
			// endless loop in case of a permanent problem, the retry counter is set to a low
			// value.
			$selectionRetries = 3;
			while (true)
			{
				// Generate a new unique visitor ID
				$ps = $db->Prepare('SELECT COUNT(*) FROM "#visitor" WHERE "VisitorId" = ?');
				$retries = 20;
				$minId = 1;
				$maxId = 2147483647;   // 2^31 - 1
				do
				{
					// Find a new random number to start with
					$tmpNewId = mt_rand($minId, $maxId);
					if ($retries == 1)
					{
						// At the very last try, scan the entire defined number range
						// (In a test, this has first occured at 99% ID space usage.)
						$tmpNewId = $minId;
					}

					$count = $db->QueryValue($ps, $tmpNewId);

					if ($count > 0 && $retries <= 3)
					{
						// Near the end, try scanning for a free ID from the random number upwards.
						// In tests, this has proven to be an efficient method to find free IDs
						// when only few free IDs are available. This query does not take
						// considerably longer than a single ID try.
						// (In a test, this has first occured at ~70% ID space usage.)
						$tmpNewId = $db->QueryValue(
							'SELECT MIN("VisitorId" + 1) AS "FreeId"
							FROM "#visitor" v1
							WHERE NOT EXISTS (SELECT * FROM "#visitor" v2 WHERE v2."VisitorId" = v1."VisitorId" + 1)
								AND "VisitorId" >= ?',
							$tmpNewId);
						if ($tmpNewId <= $maxId)
						{
							// The new ID is within the defined range, accept it
							$count = 0;
						}
					}

					$retries--;
				}
				while ($count > 0 && $retries > 0);

				if ($count > 0)
				{
					throw new Exception('No free visitor ID available.');
				}

				try
				{
					$db->Execute('INSERT INTO "#visitor" ("VisitorId") VALUES (?)', $tmpNewId);
				}
				catch (UnbDatabaseException $ex)
				{
					if ($db->IsConstraintError())
					{
						if (strpos($db->GetErrorMessage(), 'PRIMARY') !== false && $selectionRetries-- > 0)
							// Repeat the ID selection, see above.
							continue;
					}
					throw $ex;
				}

				// Leave the loop, the selected ID was still free.
				break;
			}

			$db->Commit();
			return $tmpNewId;
		}
		catch (Exception $ex)
		{
			$db->RollBack();
			throw $ex;
		}
	}

	// Merges all records of two visitors, keeping the visitor ID 1.
	//
	private function MergeVisitors($visitorId1, $visitorId2)
	{
		$db = UnbDatabase::GetInstance();
		$db->StartTransaction();
		try
		{
			// Move all sessions from visitor2 to visitor1
			$sValues = array(
				'VisitorId' => $visitorId1
				);
			$db->Update('session', $sValues, '"VisitorId" = ?', $visitorId2);

			// Merge custom visitor variables
			$visitor1 = $db->QueryRecord('SELECT * FROM "#visitor" WHERE "VisitorId" = ?', $visitorId1);
			$visitor2 = $db->QueryRecord('SELECT * FROM "#visitor" WHERE "VisitorId" = ?', $visitorId2);

			$confVisitorVars = self::GetConfVisitorVars();
			$values = array();
			foreach ($confVisitorVars as $name => $type)
			{
				if (!isset($visitor1['vv_' . $name]) && isset($visitor2['vv_' . $name]))
					$values['vv_' . $name] = $visitor2['vv_' . $name];
				else if (isset($visitor1['vv_' . $name]) && isset($visitor2['vv_' . $name]))
					$values['vv_' . $name] = max($visitor1['vv_' . $name], $visitor2['vv_' . $name]);
			}
			if (count($values))
			{
				// Something needs to be saved
				$db->Update('visitor', $values, '"VisitorId" = ?', $visitorId1);
			}

			// Delete visitorId2
			$db->Delete('visitor', '"VisitorId" = ?', $visitorId2);
			
			$db->Commit();
		}
		catch (Exception $ex)
		{
			$db->RollBack();
			throw $ex;
		}
	}

	
	// -------------------------------  Database: Session functions  ------------------------------


	// Finds a session by its session ID and gets its data.
	//
	// sessionId = (int) Session ID
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbDatabaseException
	//
	// Returns an array of the session data if successful, {{null}} otherwise.
	//
	private function FindSessionById($sessionId)
	{
		if (!is_int($sessionId))
			throw new InvalidArgumentException('Invalid argument type: $sessionId');
		if ($sessionId <= 0)
			throw new InvalidArgumentException('Argument out of range: $sessionId');

		$db = UnbDatabase::GetInstance();
		$rec = $db->QueryRecord('SELECT * FROM "#session" WHERE "SessionId" = ?', $sessionId);
		if (!$rec) return null;
		return $rec;
	}
	
	// Gets a value indicating whether the specified session ID exists.
	//
	// Returns (bool) {{true}} if the session ID exists, {{false}} otherwise.
	//
	private function ExistsSessionId($sessionId)
	{
		if (!is_int($sessionId))
			throw new InvalidArgumentException('Invalid argument type: $sessionId');
		if ($sessionId <= 0)
			throw new InvalidArgumentException('Argument out of range: $sessionId');

		$db = UnbDatabase::GetInstance();
		$rec = $db->QueryRecord('SELECT "SessionId" FROM "#session" WHERE "SessionId" = ?', $sessionId);
		return $rec ? true : false;
	}
	
	// Finds all current sessions for the visitor.
	//
	// Returns an array of all sessions with their session data if successful, {{null}} otherwise.
	//
	private function GetCurrentSessionsForVisitor($visitorId)
	{
		if (!is_int($visitorId))
			throw new InvalidArgumentException('Invalid argument type: $visitorId');
		if ($visitorId <= 0)
			throw new InvalidArgumentException('Argument out of range: $visitorId');

		$db = UnbDatabase::GetInstance();
		$recs = $db->Query(
			'SELECT *
			FROM "#session"
			WHERE "VisitorId" = ?
				AND "SessionUpdateTime" > ?
			ORDER BY "SessionUpdateTime" DESC',
			array(
				$visitorId,
				UnbDatabase::ConvertDateTime(time() - intval(UnbConfiguration::Get('session.maxidle', '30')) * 60)
			));
		if (!$recs) return null;
		return $recs;
	}
	
	// Finds an existing session that is similar to the passed data which could be used to create
	// a new session.
	//
	// Only sessions with a minimum level of similarity are considered. Among those, if any, only
	// the most closely matching session is returned.
	//
	private function FindSimilarSession($dbData)
	{
		// Read all active sessions that indicate no cookies support, and all sessions that have
		// been started within the last 10 seconds. The first case includes regular requests from
		// clients with no cookie support. The second case handles very quick subsequent requests
		// (like script and noscript request) where the cookie from the first reply has not been
		// considered yet for the second request. This usually only happens in a very short
		// timeframe.
		$db = UnbDatabase::GetInstance();
		
		if ($this->trackDntLevel >= 2)
		{
			// No visitor cookie is set even if the browser would accept it. So we need to perform
			// the cookie-less session recognition for the full time, but we can use the DNT value
			// to narrow down the choices.
			$recs = $db->Query(
				'SELECT *
				FROM "#session"
				WHERE "DNT" = 1 AND "SessionUpdateTime" > ?
				ORDER BY "SessionUpdateTime"',
				array(
					UnbDatabase::ConvertDateTime(time() - intval(UnbConfiguration::Get('session.maxidle', '30')) * 60),
				));
		}
		else
		{
			$recs = $db->Query(
				'SELECT *
				FROM "#session"
				WHERE (("Cookies" = 0 OR "Cookies" IS NULL) AND "SessionUpdateTime" > ?)
					OR "SessionStartTime" > ?
				ORDER BY "SessionUpdateTime"',
				array(
					UnbDatabase::ConvertDateTime(time() - intval(UnbConfiguration::Get('session.maxidle', '30')) * 60),
					UnbDatabase::ConvertDateTime(time() - 10)
				));
		}
		if (!$recs) return null;
		
		$minSimilarity = 10;
		$bestSession = null;
		$bestSimilarity = 0;
		UnbDebug::Trace('    considering ' . count($recs) . ' sessions for similarity');
		foreach ($recs as $session)
		{
			$similarity = $this->ComputeSimilarity($dbData, $session);
			if ($similarity > $minSimilarity && $similarity > $bestSimilarity)
			{
				$bestSimilarity = $similarity;
				$bestSession = $session;
			}
		}
		if (isSet($bestSession))
		{
			UnbDebug::Trace('    bestSession: SessionId=' . $bestSession['SessionId'] . ', SessionStartTime=' . $bestSession['SessionStartTime'] . ', SessionUpdateTime=' . $bestSession['SessionUpdateTime'] . ', Cookies=' . $bestSession['Cookies']);
		}
		return $bestSession;
	}
	
	// Computes the similarity of two sessions.
	//
	// session1 = (array) New session data to find an existing match for
	// session2 = (array) Existing (older) session data to compare
	//
	private function ComputeSimilarity($session1, $session2)
	{
		$sum = 0;
		$sum += ($session1['IPAddress'] == $session2['IPAddress'] ? 4 : -4);
		$sum += ($session1['UserAgentString'] == $session2['UserAgentString'] ? 4 : -4);
		#$sum += ($session1['BrowserText'] == $session2['BrowserText'] &&
		#	$session1['OperatingSystemText'] == $session2['OperatingSystemText'] ? 4 : -4);
		$sum += ($session1['LanguagesString'] == $session2['LanguagesString'] ? 4 : -4);

		// Only compare data collected with JavaScript if both sessions indicate JS support
		if ($session1['JavaScript'] && $session2['JavaScript'])
		{
			$sum += ($session1['ScreenWidth'] == $session2['ScreenWidth'] &&
				$session1['ScreenHeight'] == $session2['ScreenHeight'] &&
				$session1['ScreenColors'] == $session2['ScreenColors'] ? 2 : -3);
			$sum += ($session1['ScreenAvailWidth'] == $session2['ScreenAvailWidth'] &&
				$session1['ScreenAvailHeight'] == $session2['ScreenAvailHeight'] ? 2 : -3);
			$sum += ($session1['ViewportWidth'] == $session2['ViewportWidth'] &&
				$session1['ViewportHeight'] == $session2['ViewportHeight'] ? 3 : -1);
			$sum += ($session1['UtcOffset'] == $session2['UtcOffset'] ? 3 : -3);
		}
		return $sum;
	}

	// Converts the raw reported data to field values to be used in the database.
	//
	// This is used by CreateSession and can also be used to create similar data to pass to
	// FindSimilarSession.
	//
	// Returns an array of the prepared session data (without SessionId).
	//
	private function PrepareSessionData($rawData, $now, $visitorId, $newVisitor)
	{
		$values = array(
			'SessionStartTime' => UnbDatabase::ConvertDateTime($now),
			'SessionUpdateTime' => UnbDatabase::ConvertDateTime($now),
			'VisitorId' => $visitorId,
			'NewVisitor' => $newVisitor ? 1 : 0,
			);

		$values = array_merge(
			$values,
			$this->GetLocationSessionValues($rawData),
			$this->GetReferrerSessionValues($rawData),
			$this->GetBrowserSessionValues($rawData));

		$values['UserAgentString'] = $_SERVER['HTTP_USER_AGENT'];
		$values['LanguagesString'] = $_SERVER['HTTP_ACCEPT_LANGUAGE'];

		return $values;
	}

	// Creates a new session in the database.
	//
	// Returns an array of the session data.
	//
	private function CreateSession($rawData, $now, $visitorId, $newVisitor)
	{
		$values = $this->PrepareSessionData($rawData, $now, $visitorId, $newVisitor);
		$values['SessionId'] = new UnbDbAutoValue('session');
		
		// Add custom session variables
		$confSessionVars = self::GetConfSessionVars();
		foreach ($confSessionVars as $name => $type)
		{
			if (isset($rawData['sv_' . $name]))
			{
				if ($type == 'integer')
				{
					$values['sv_' . $name] = intval($rawData['sv_' . $name]);
				}
				else if ($type == 'text')
				{
					$values['sv_' . $name] = self::GetTextId($rawData['sv_' . $name]);
				}
			}
		}
		
		// Also add/update custom visitor variables if set
		$confVisitorVars = self::GetConfVisitorVars();
		$visitorValues = array();
		foreach ($confVisitorVars as $name => $type)
		{
			if (isset($rawData['vv_' . $name]))
			{
				if ($type == 'integer')
				{
					$visitorValues['vv_' . $name] = intval($rawData['vv_' . $name]);
				}
				else if ($type == 'text')
				{
					$visitorValues['vv_' . $name] = self::GetTextId($rawData['vv_' . $name]);
				}
			}
		}
		
		$db = UnbDatabase::GetInstance();
		$db->StartTransaction();
		try
		{
			$db->Insert('session', $values);
			$values['SessionId'] = $db->GetLastAutoValue();

			if (count($visitorValues))
			{
				$db->Update('visitor', $visitorValues, '"VisitorId" = ?', $visitorId);
			}

			$db->Commit();
		}
		catch (Exception $ex)
		{
			$db->RollBack();
			throw $ex;
		}

		return $values;
	}
	
	// Updates a session in the database with new data from the client.
	//
	// session = (array) Session data that is already stored in the database and shall be updated
	// rawData = (array) New data reported from the client
	// now = (int) Common request timestamp
	//
	public function UpdateSession($session, $rawData, $now)
	{
		$clearExitPages = isset($session['SessionDuration']);
		$db = UnbDatabase::GetInstance();
		
		$values = array(
			'SessionUpdateTime' => UnbDatabase::ConvertDateTime($now),
			'SessionDuration' => null,   // Will be recomputed later
			);

		// Only those values can be updated that could possibly change throughout a visitor's
		// session. That includes JavaScript and cookies support as well as any information that
		// is only available with JavaScript support. Things like IP address or browser identifier
		// are available from the first request on, so they don't need to be updated later. Also,
		// the referrer and campaign name cannot be considered on any later request than the first
		// one. (The second one is also okay if it comes with Js data while the first did not.)
		
		$locationValues = $this->GetLocationSessionValues($rawData);
		$refValues = $this->GetReferrerSessionValues($rawData);
		$browserValues = $this->GetBrowserSessionValues($rawData);
		
		if ($browserValues['JavaScript'])
		{
			// JavaScript is available for this request, so the data collected through JavaScript
			// can be regarded here.
			$values['JavaScript'] = 1;

			// OR the supported features and plugins in case they've been available only some of the time.
			$values['Flash'] = max($session['Flash'], $browserValues['Flash']);
			$values['Java'] = max($session['Java'], $browserValues['Java']);
			$values['Quicktime'] = max($session['Quicktime'], $browserValues['Quicktime']);
			$values['Silverlight'] = max($session['Silverlight'], $browserValues['Silverlight']);
			$values['WindowsMedia'] = max($session['WindowsMedia'], $browserValues['WindowsMedia']);
			$values['AdBlock'] = max($session['AdBlock'], $browserValues['AdBlock']);

			// Only set time/screen information if it was unset, to keep the data from the earliest request.
			if (!isset($session['UtcOffset']) && isset($browserValues['UtcOffset']))
			{
				$values['UtcOffset'] = $browserValues['UtcOffset'];

				// Update the location data once, now that the time zone has come to our knowledge.
				if (isset($locationValues['CountryCode']))
				{
					// Update the complete set of location values at once, not each one individually!
					$values['CountryCode'] = $locationValues['CountryCode'];
					$values['RegionText'] = $locationValues['RegionText'];
					$values['CityText'] = $locationValues['CityText'];
					$values['Lat'] = $locationValues['Lat'];
					$values['Lng'] = $locationValues['Lng'];
				}
			}
			if (!$session['ScreenWidth'] && $browserValues['ScreenWidth'])
			{
				$values['ScreenWidth'] = $browserValues['ScreenWidth'];
			}
			if (!$session['ScreenHeight'] && $browserValues['ScreenHeight'])
			{
				$values['ScreenHeight'] = $browserValues['ScreenHeight'];
			}
			if (!$session['ScreenAvailWidth'] && $browserValues['ScreenAvailWidth'])
			{
				$values['ScreenAvailWidth'] = $browserValues['ScreenAvailWidth'];
			}
			if (!$session['ScreenAvailHeight'] && $browserValues['ScreenAvailHeight'])
			{
				$values['ScreenAvailHeight'] = $browserValues['ScreenAvailHeight'];
			}
			if (!$session['ScreenColors'] && $browserValues['ScreenColors'])
			{
				$values['ScreenColors'] = $browserValues['ScreenColors'];
			}
			if (!$session['ViewportWidth'] && $browserValues['ViewportWidth'])
			{
				$values['ViewportWidth'] = $browserValues['ViewportWidth'];
			}
			if (!$session['ViewportHeight'] && $browserValues['ViewportHeight'])
			{
				$values['ViewportHeight'] = $browserValues['ViewportHeight'];
			}

			// The following values can only be written to the session if no Js data has been
			// available before, and now it is. There can only be a single first request with
			// JavaScript data in a session.
			if (!$session['JavaScript'])
			{
				// These fields may only be set with the first pageview in a session. Otherwise,
				// the referrer field is overwritten by a referring page from the middle of our own
				// session, if the user enables JavaScript while browsing the site.
				
				// Determine how many pageviews already exist for this session
				$pageviewCount = $db->QueryValue(
					'SELECT COUNT(*) FROM "#pageview" WHERE "SessionId" = ?',
					intval($session['SessionId']));
				if (isset($rawData[self::key_pageviewId]))
				{
					// This is a pageview update request. UpdateSession is called for the first
					// pageview of the session, if a single pageview exists in the session.
					$isFirst = $pageviewCount == 1;
				}
				else
				{
					// This is a pageview enter request. UpdateSession is called for the first
					// pageview of the session, if no pageview exists in the session.
					$isFirst = $pageviewCount == 0;
				}
				
				if ($isFirst)
				{
					if (!isset($session['SearchKeywordsText']) && isset($refValues['SearchKeywordsText']))
					{
						$values['SearchKeywordsText'] = $refValues['SearchKeywordsText'];
					}
					if (!isset($session['SearchEngineText']) && isset($refValues['SearchEngineText']))
					{
						$values['SearchEngineText'] = $refValues['SearchEngineText'];
						$values['ReferrerUrlPathText'] = null;
					}
					if (!isset($session['SearchImageUrlText']) && isset($refValues['SearchImageUrlText']))
					{
						$values['SearchImageUrlText'] = $refValues['SearchImageUrlText'];
					}
					if (!isset($session['SearchResultPage']) && isset($refValues['SearchResultPage']))
					{
						$values['SearchResultPage'] = $refValues['SearchResultPage'];
					}

					if (!isset($session['ReferrerSsl']) && isset($refValues['ReferrerSsl']))
					{
						$values['ReferrerSsl'] = $refValues['ReferrerSsl'];
					}
					if (!isset($session['ReferrerHostText']) && isset($refValues['ReferrerHostText']))
					{
						$values['ReferrerHostText'] = $refValues['ReferrerHostText'];
					}
					if (!isset($session['ReferrerUrlPathText']) && isset($refValues['ReferrerUrlPathText']))
					{
						$values['ReferrerUrlPathText'] = $refValues['ReferrerUrlPathText'];
					}

					if (!$session['CampaignText'] && $refValues['CampaignText'])
					{
						$values['CampaignText'] = $refValues['CampaignText'];
						UnbDebug::Trace('      Campaign: ' . $rawData[self::key_campaign]);
					}
				}
			}
		}

		// Always overwrite cookies support because a first information could have been wrong.
		if (strlen($rawData[self::key_cookies]))
		{
			$values['Cookies'] = $rawData[self::key_cookies] ? 1 : 0;
		}
		
		// Update custom session variables
		$confSessionVars = self::GetConfSessionVars();
		foreach ($confSessionVars as $name => $type)
		{
			if (isset($rawData['sv_' . $name]))
			{
				if ($type == 'integer')
				{
					$values['sv_' . $name] = intval($rawData['sv_' . $name]);
				}
				else if ($type == 'text')
				{
					$values['sv_' . $name] = self::GetTextId($rawData['sv_' . $name]);
				}
			}
		}
		
		// Also update custom visitor variables if set
		$confVisitorVars = self::GetConfVisitorVars();
		$visitorValues = array();
		foreach ($confVisitorVars as $name => $type)
		{
			if (isset($rawData['vv_' . $name]))
			{
				if ($type == 'integer')
				{
					$visitorValues['vv_' . $name] = intval($rawData['vv_' . $name]);
				}
				else if ($type == 'text')
				{
					$visitorValues['vv_' . $name] = self::GetTextId($rawData['vv_' . $name]);
				}
			}
		}
		
		$db->StartTransaction();
		try
		{
			$db->Update('session', $values, '"SessionId" = ?', intval($session['SessionId']));
			
			if ($clearExitPages)
			{
				// Clear exit flag from all pageviews that may have been left
				$db->Execute(
					'UPDATE "#pageview" SET "ExitPage" = NULL WHERE "SessionId" = ?',
					intval($session['SessionId']));
			}
			
			if (count($visitorValues))
			{
				$db->Update('visitor', $visitorValues, '"VisitorId" = ?', intval($session['VisitorId']));
			}

			$db->Commit();
		}
		catch (Exception $ex)
		{
			$db->RollBack();
			throw $ex;
		}
	}
	
	// Merges two sessions into one, updating all references and combining the data.
	//
	private function MergeSessions($session1, $session2)
	{
		if ($session1['VisitorId'] != $session2['VisitorId'])
			throw new Exception('Only sessions of the same visitor can be merged.');
		if ($session1['NewVisitor'] != $session2['NewVisitor'])
			throw new Exception('The session value NewVisitor cannot be merged.');
		
		// session1 will be kept. Merge in all relevant data from session2.
		$values['SessionStartTime'] = min($session1['SessionStartTime'], $session2['SessionStartTime']);
		$values['SessionDuration'] = null;   // Will be recomputed later
		$values['SessionUpdateTime'] = max($session1['SessionUpdateTime'], $session2['SessionUpdateTime']);
		$values['ReferrerSsl'] = $this->MergeSessionHelper($session1, $session2, 'ReferrerSsl', 1, true);
		$values['ReferrerHostText'] = $this->MergeSessionHelper($session1, $session2, 'ReferrerHostText', 1, true);
		$values['ReferrerUrlPathText'] = $this->MergeSessionHelper($session1, $session2, 'ReferrerUrlPathText', 1, true);
		$values['SearchEngineText'] = $this->MergeSessionHelper($session1, $session2, 'SearchEngineText', 1, true);
		$values['SearchKeywordsText'] = $this->MergeSessionHelper($session1, $session2, 'SearchKeywordsText', 1, true);
		$values['SearchImageUrlText'] = $this->MergeSessionHelper($session1, $session2, 'SearchImageUrlText', 1, true);
		$values['SearchResultPage'] = $this->MergeSessionHelper($session1, $session2, 'SearchResultPage', 1, true);
		$values['CampaignText'] = $this->MergeSessionHelper($session1, $session2, 'CampaignText', 1, true);

		// Always take all location values from the same source, don't select individually!
		// I.e., if a source is selected and contains null while the other has a value, keep null.
		// Decide by the country code, then by latitude. If all are equally set or unset, keep the
		// values from session1.
		
		// First try to select the single session which has CountryCode set.
		if (isset($session1['CountryCode']) && !isset($session2['CountryCode'])) $locationSession = &$session1;
		else if (!isset($session1['CountryCode']) && isset($session2['CountryCode'])) $locationSession = &$session2;
		// If both are unset, it doesn't matter (nothing will be set on either side).
		// If both are set, prefer the session with JavaScript (allows higher precision).
		else if ($session1['JavaScript'] && !$session2['JavaScript']) $locationSession = &$session1;
		else if (!$session1['JavaScript'] && $session2['JavaScript']) $locationSession = &$session2;
		// If there's no difference, prefer the session with Lat set (higher resolution).
		else if (isset($session1['Lat']) && !isset($session2['Lat'])) $locationSession = &$session1;
		else if (!isset($session1['Lat']) && isset($session2['Lat'])) $locationSession = &$session2;
		// Can't decide (or don't need to because both are equal), keep as is.
		else $locationSession = &$session1;

		$values['CountryCode'] = $locationSession['CountryCode'];
		$values['RegionText'] = $locationSession['RegionText'];
		$values['CityText'] = $locationSession['CityText'];
		$values['Lat'] = $locationSession['Lat'];
		$values['Lng'] = $locationSession['Lng'];

		$values['UtcOffset'] = $this->MergeSessionHelper($session1, $session2, 'UtcOffset', 0, true);
		$values['JavaScript'] = max($session1['JavaScript'], $session2['JavaScript']);
		$values['Cookie'] = max($session1['Cookie'], $session2['Cookie']);
		$values['Flash'] = max($session1['Flash'], $session2['Flash']);
		$values['Java'] = max($session1['Java'], $session2['Java']);
		$values['Quicktime'] = max($session1['Quicktime'], $session2['Quicktime']);
		$values['Silverlight'] = max($session1['Silverlight'], $session2['Silverlight']);
		$values['WindowsMedia'] = max($session1['WindowsMedia'], $session2['WindowsMedia']);
		$values['AdBlock'] = max($session1['AdBlock'], $session2['AdBlock']);
		$values['ScreenWidth'] = $this->MergeSessionHelper($session1, $session2, 'ScreenWidth', 1, true);
		$values['ScreenHeight'] = $this->MergeSessionHelper($session1, $session2, 'ScreenHeight', 1, true);
		$values['ScreenAvailWidth'] = $this->MergeSessionHelper($session1, $session2, 'ScreenAvailWidth', 1, true);
		$values['ScreenAvailHeight'] = $this->MergeSessionHelper($session1, $session2, 'ScreenAvailHeight', 1, true);
		$values['ScreenColors'] = $this->MergeSessionHelper($session1, $session2, 'ScreenColors', 1, true);
		$values['ViewportWidth'] = $this->MergeSessionHelper($session1, $session2, 'ViewportWidth', 1, true);
		$values['ViewportHeight'] = $this->MergeSessionHelper($session1, $session2, 'ViewportHeight', 1, true);
		
		// Merge custom session variables
		$confSessionVars = self::GetConfSessionVars();
		foreach ($confSessionVars as $name => $type)
		{
			$values['sv_' . $name] = $this->MergeSessionHelper($session1, $session2, 'sv_' . $name, 1, true);
		}
		
		$db = UnbDatabase::GetInstance();
		$db->StartTransaction();
		try
		{
			// Update session1 with all new data
			$db->Update('session', $values, '"SessionId" = ?', intval($session1['SessionId']));

			// Move all pageviews from session2 to session1
			$pvValues = array(
				'SessionId' => intval($session1['SessionId'])
				);
			$db->Update('pageview', $pvValues, '"SessionId" = ?', intval($session2['SessionId']));
			
			// Delete session2
			$db->Delete('session', '"SessionId" = ?', intval($session2['SessionId']));

			$db->Commit();
		}
		catch (Exception $ex)
		{
			$db->RollBack();
			throw $ex;
		}
	}

	// Purges temporary additional data from expired sessions.
	//
	public function PurgeTempSessionData()
	{
		$db = UnbDatabase::GetInstance();
		$db->StartTransaction();
		try
		{
			$db->Update(
				'session',
				array(
					'UserAgentString' => null,
					'LanguagesString' => null),
				'"SessionUpdateTime" < ?
					AND ("UserAgentString" IS NOT NULL OR "LanguagesString" IS NOT NULL)',
				array(
					UnbDatabase::ConvertDateTime(time() - intval(UnbConfiguration::Get('session.maxidle', '30')) * 60)
				));

			$db->Commit();
		}
		catch (Exception $ex)
		{
			$db->RollBack();
			throw $ex;
		}
	}

	
	// ------------------------------  Database: Pageview functions  ------------------------------
	
	
	// Finds a pageview by its pageview ID.
	//
	// pageviewId = (int) Pageview ID
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbDatabaseException
	//
	// Returns an array of the pageview data if successful, {{null}} otherwise.
	//
	private function FindPageviewById($pageviewId)
	{
		if (!is_int($pageviewId))
			throw new InvalidArgumentException('Invalid argument type: $pageviewId');
		if ($pageviewId < 0)
			throw new InvalidArgumentException('Argument out of range: $pageviewId');

		$db = UnbDatabase::GetInstance();
		$rec = $db->QueryRecord('SELECT * FROM "#pageview" WHERE "PageviewId" = ?', $pageviewId);
		if (!$rec) return null;
		return $rec;
	}

	// Gets a value indicating whether the specified pageview ID exists.
	//
	// Returns (bool) {{true}} if the pageview ID exists, {{false}} otherwise.
	//
	private function ExistsPageviewId($pageviewId)
	{
		if (!is_int($pageviewId))
			throw new InvalidArgumentException('Invalid argument type: $pageviewId');
		if ($pageviewId < 0)
			throw new InvalidArgumentException('Argument out of range: $pageviewId');

		$db = UnbDatabase::GetInstance();
		$rec = $db->QueryRecord('SELECT "PageviewId" FROM "#pageview" WHERE "PageviewId" = ?', $pageviewId);
		return $rec ? true : false;
	}
	
	// Finds existing pageviews that are a duplicate of the passed data which could be used to
	// create a new pageview.
	//
	private function FindDuplicatePageviews($dbData)
	{
		// Read all pageviews of similar time from the same session, but not the passed pageview
		// (if it already exists in the database)
		$maxOffset = 60;   // allowed offset in seconds
		$db = UnbDatabase::GetInstance();
		$recs = $db->Query(
			'SELECT *
			FROM "#pageview"
			WHERE "SessionId" = ?
				AND "PageviewId" <> ?
				AND "PageviewTime" >= ?
				AND "PageviewTime" <= ?
				AND "CampaignRedirect" = ?
			ORDER BY "PageviewId" DESC',
			array(
				intval($dbData['SessionId']),
				intval($dbData['PageviewId']),
				UnbDatabase::ConvertDateTime(time() - $maxOffset),
				UnbDatabase::ConvertDateTime(time() + $maxOffset),
				intval($dbData['CampaignRedirect'])
			));
		if (!$recs) return array();
		
		$dupes = array();
		foreach ($recs as $pageview)
		{
			// We're iterating all pageviews in the allowed time in reverse order of creation.
			// Collect all duplicates as long as the page URL matches. As soon as the page URL is
			// different, stop considering pageviews as duplicate, because even if we find an
			// older pageview with the same URL, the other intermediate pageviews indicate that
			// there has been a legitimate navigation meanwhile and the new request must be a
			// different pageview, i.e. not a duplicate.
			if ($pageview['Ssl'] == $dbData['Ssl'] &&
				$pageview['DomainText'] == $dbData['DomainText'] &&
				$pageview['UrlPathText'] == $dbData['UrlPathText'] &&
				$pageview['UrlQueryText'] == $dbData['UrlQueryText'])
			{
				if (!$dbData['JavaScriptData'] &&   // This is a no-script request
					$pageview['JavaScriptData'] &&   // Comparing with a script pageview
					$pageview['PageLoadTime'] > 0)   // Comparing with a fully loaded pageview
				{
					// A noscript request cannot update a pageview which is already marked as fully
					// loaded, i.e. has already received the imgns response from the <noscript>
					// image.
				}
				else if ($dbData['JavaScriptData'] &&   // This is a JavaScript request
					$pageview['JavaScriptData'] &&   // Comparing with a script pageview
					isset($pageview['PageLoadTime']) &&   // Comparing with a virtual pageview
					$pageview['PageLoadTime'] == 0)
				{
					// Virtual pageviews cannot be duplicates, they often have the same URL but
					// were created to indicate the display of different content.
				}
				else if ($dbData['JavaScriptData'] &&   // This is a JavaScript request
					$pageview['JavaScriptData'] &&   // Comparing with a script pageview
					$pageview['Prerender'] !== $dbData['Prerender'])   // Prerender differs
				{
					// Don't overwrite anything with a different Prerender status as those may be
					// purged in the future. Also, both of these pageviews are reported by
					// JavaScript, which usually only happens once per actual pageview, so don't
					// consider this a duplicate.
				}
				else
				{
					$dupes[] = $pageview;
				}
			}
			else
			{
				break;
			}
		}
		return $dupes;
	}
	
	// Creates a new pageview in the database.
	//
	// Returns an array of the pageview data.
	//
	private function CreatePageview($rawData, $now, $sessionId, $entryPage)
	{
		$values = $this->PreparePageviewData($rawData, $now, $sessionId, $entryPage);
		$values['PageviewId'] = new UnbDbAutoValue('pageview');
		if (!isset($values['Prerender']))
		{
			$values['Prerender'] = 0;   // Must be set
		}

		// Add custom pageview variables
		$confPageviewVars = self::GetConfPageviewVars();
		foreach ($confPageviewVars as $name => $type)
		{
			if (isset($rawData['pv_' . $name]))
			{
				if ($type == 'integer')
				{
					$values['pv_' . $name] = intval($rawData['pv_' . $name]);
				}
				else if ($type == 'text')
				{
					$values['pv_' . $name] = self::GetTextId($rawData['pv_' . $name]);
				}
			}
		}
		
		$db = UnbDatabase::GetInstance();
		$db->StartTransaction();
		try
		{
			$db->Insert('pageview', $values);
			$values['PageviewId'] = $db->GetLastAutoValue();

			$db->Commit();
		}
		catch (Exception $ex)
		{
			$db->RollBack();
			throw $ex;
		}
		return $values;
	}

	// Updates a pageview in the database with new data from the client.
	//
	// pageview = (array) Pageview data that is already stored in the database and shall be updated
	// rawData = (array) New data reported from the client
	// now = (int) Common request timestamp
	//
	public function UpdatePageview($pageview, $rawData, $now)
	{
		$pageviewId = intval($pageview['PageviewId']);
		if ($pageviewId <= 0)
			throw new InvalidArgumentException('Argument out of range: $pageview[\'PageviewId\']');
		
		$values = array();
		
		// If updating a non-js pageview with Js data, don't set the update time for less than 5 seconds
		if (!$pageview['JavaScriptData'] && $rawData['javascript'])
		{
			// Updating non-js with js
			$values['JavaScriptData'] = 1;
			if ($now - UnbDatabase::ConvertDateTime($pageview['PageviewUpdateTime']) >= 5)
			{
				// More than 5 seconds ago
				$values['PageviewUpdateTime'] = UnbDatabase::ConvertDateTime($now);
			}

			// The OriginalPageviewId cannot be updated after Js data has been available because it
			// would always overwrite the previous NULL value when reloading the page normally.
			if (intval($rawData[self::key_origPageviewId]) > 0)
			{
				$values['OriginalPageviewId'] = intval($rawData[self::key_origPageviewId]);
			}
		}
		else
		{
			$values['PageviewUpdateTime'] = UnbDatabase::ConvertDateTime($now);
		}

		if (!$pageview['UrlHashText'] && isset($rawData[self::key_pageUrl]))
		{
			// Parse page URL into its elements
			$pageUrl = parse_url($rawData[self::key_pageUrl]);

			// Only the hash value may be updated, if it is new to the pageview record
			$values['UrlHashText'] = Tomo::GetTextId($pageUrl['fragment']);
		}

		if (strlen($rawData[self::key_pageTitle]))
		{
			$values['PageTitleText'] = Tomo::GetTextId($rawData[self::key_pageTitle]);
		}
		if (strlen($rawData[self::key_pageLanguage]))
		{
			$values['PageLanguage'] = $rawData[self::key_pageLanguage];
		}

		if ($rawData[self::key_pageLoadTime] != 0)
		{
			$values['PageLoadTime'] = intval($rawData[self::key_pageLoadTime]);
		}
		if (isset($rawData[self::key_activityTimespan]))
		{
			$activityTimespan = intval($rawData[self::key_activityTimespan]);
			// Set the rounding threshold to 0.8, i.e. count a full second after 800 ms of it.
			$activityTimespan = intval(($activityTimespan + 200) / 1000);
			$pageviewTime = UnbDatabase::ConvertDateTime($pageview['PageviewTime']);
			if ($activityTimespan > $now - $pageviewTime + 2)
			{
				UnbDebug::Trace('  Reported activity timespan is larger than absolute pageview lifetime. Limit value.');
				$activityTimespan = $now - $pageviewTime;
			}
			$values['ActivityTimespan'] = $activityTimespan;
		}
		// Any positive activity timespan requires prerendering to be left, so clear it even if not
		// explicitly indicated by the client (a request may have been lost etc.)
		if (isset($rawData[self::key_prerender]) || $rawData[self::key_activityTimespan] > 0)
		{
			$values['Prerender'] = $rawData[self::key_prerender] ? 1 : 0;
		}
		
		if (strlen($rawData[self::key_campaign]) && $rawData['javascript'])
		{
			$values['CampaignRedirect'] = 1;
		}

		// Update custom pageview variables
		$confPageviewVars = self::GetConfPageviewVars();
		foreach ($confPageviewVars as $name => $type)
		{
			if (isset($rawData['pv_' . $name]))
			{
				if ($type == 'integer')
				{
					$values['pv_' . $name] = intval($rawData['pv_' . $name]);
				}
				else if ($type == 'text')
				{
					$values['pv_' . $name] = self::GetTextId($rawData['pv_' . $name]);
				}
			}
		}
		
		if (!count($values)) return;   // Nothing to do

		$db = UnbDatabase::GetInstance();
		$db->StartTransaction();
		try
		{
			$db->Update('pageview', $values, '"PageviewId" = ?', $pageviewId);

			$db->Commit();
		}
		catch (Exception $ex)
		{
			$db->RollBack();
			throw $ex;
		}
	}
	
	// Duplicates a pageview.
	//
	private function DuplicatePageview($pageview, $newRawData, $now)
	{
		$values = array();
		foreach ($pageview as $key => $value)
		{
			// Copy all pageview fields except custom variables
			if (substr($key, 0, 3) != 'pv_')
				$values[$key] = $value;
		}
		
		$values['PageviewId'] = new UnbDbAutoValue('pageview');
		$values['PageviewTime'] = UnbDatabase::ConvertDateTime($now);
		$values['PageviewUpdateTime'] = UnbDatabase::ConvertDateTime($now);
		if (strlen($newRawData[self::key_pageTitle]))
		{
			$values['PageTitleText'] = Tomo::GetTextId($newRawData[self::key_pageTitle]);
		}
		if (strlen($newRawData[self::key_pageLanguage]))
		{
			$values['PageLanguage'] = $newRawData[self::key_pageLanguage];
		}
		$values['ActivityTimespan'] = null;
		$values['PageLoadTime'] = 0;
		$values['EntryPage'] = 0;
		$values['JavaScriptData'] = 1;
		$values['CampaignRedirect'] = 0;
		$values['Prerender'] = 0;
		$values['OriginalPageviewId'] = null;

		// Update custom pageview variables
		$confPageviewVars = self::GetConfPageviewVars();
		foreach ($confPageviewVars as $name => $type)
		{
			if (isset($newRawData['pv_' . $name]))
			{
				if ($type == 'integer')
				{
					$values['pv_' . $name] = intval($newRawData['pv_' . $name]);
				}
				else if ($type == 'text')
				{
					$values['pv_' . $name] = self::GetTextId($newRawData['pv_' . $name]);
				}
			}
		}
		
		$db = UnbDatabase::GetInstance();
		$db->StartTransaction();
		try
		{
			$db->Insert('pageview', $values);
			$values['PageviewId'] = $db->GetLastAutoValue();

			$db->Commit();
		}
		catch (Exception $ex)
		{
			$db->RollBack();
			throw $ex;
		}
		return $values;
	}

	// Merges two pageviews into one, updating all references and combining the data.
	//
	// pageview1 = (array) Pageview data to keep.
	// pageview2 = (array) Pageview data to merge into pageview1. This will be deleted from the database.
	//
	private function MergePageviews($pageview1, $pageview2)
	{
		if ($pageview1['SessionId'] != $pageview2['SessionId'])
			throw new Exception('Only pageviews of the same session can be merged.');
		
		// pageview1 will be kept. Merge in all relevant data from pageview2.
		$values['PageviewTime'] = min($pageview1['PageviewTime'], $pageview2['PageviewTime']);
		$values['PageTitleText'] = max($pageview1['PageTitleText'], $pageview2['PageTitleText']);
		$values['ActivityTimespan'] = max($pageview1['ActivityTimespan'], $pageview2['ActivityTimespan']);
		$values['JavaScriptData'] = max($pageview1['JavaScriptData'], $pageview2['JavaScriptData']);
		$values['PageLoadTime'] = max($pageview1['PageLoadTime'], $pageview2['PageLoadTime']);
		$values['EntryPage'] = max($pageview1['EntryPage'], $pageview2['EntryPage']);
		$values['CampaignRedirect'] = max($pageview1['CampaignRedirect'], $pageview2['CampaignRedirect']);
		$values['Prerender'] = max($pageview1['Prerender'], $pageview2['Prerender']);
		$values['OriginalPageviewId'] = min($pageview1['OriginalPageviewId'], $pageview2['OriginalPageviewId']);
		$values['PageviewUpdateTime'] = max($pageview1['PageviewUpdateTime'], $pageview2['PageviewUpdateTime']);
		
		// Merge custom pageview variables
		$confPageviewVars = self::GetConfPageviewVars();
		foreach ($confPageviewVars as $name => $type)
		{
			$values['pv_' . $name] = max($pageview1['pv_' . $name], $pageview2['pv_' . $name]);
		}
		
		$db = UnbDatabase::GetInstance();
		$db->StartTransaction();
		try
		{
			// Update pageview1 with all new data
			$db->Update('pageview', $values, '"PageviewId" = ?', intval($pageview1['PageviewId']));

			// Move all actions from pageview2 to pageview1
			$acValues = array(
				'PageviewId' => intval($pageview1['PageviewId'])
				);
			$db->Update('action', $acValues, '"PageviewId" = ?', intval($pageview2['PageviewId']));
			
			// Delete pageview2
			$db->Delete('pageview', '"PageviewId" = ?', intval($pageview2['PageviewId']));

			$db->Commit();
		}
		catch (Exception $ex)
		{
			$db->RollBack();
			throw $ex;
		}
	}
	
	// Sets the ExitPage flag for the last pageview of all expired sessions.
	//
	public function UpdateAllExitPages()
	{
		// Session expiry time, with a bit safety margin
		$time = time() - (intval(UnbConfiguration::Get('session.maxidle', '30')) + 1) * 60;
		$time = UnbDatabase::ConvertDateTime($time);

		$db = UnbDatabase::GetInstance();
		$db->StartTransaction();
		try
		{
			foreach ($db->Query('
				SELECT "SessionId", "SessionStartTime", "SessionUpdateTime"
				FROM "#session"
				WHERE "SessionDuration" IS NULL
					AND "SessionUpdateTime" < ?',
				$time)
				as $session)
			{
				$sessionId = intval($session['SessionId']);
				
				// Find the latest pageview update time. Should be the same as session.SessionUpdateTime
				#$lastPageviewUpdateTime = $db->QueryValue(
				#	'SELECT MAX("PageviewUpdateTime") FROM "#pageview" WHERE "SessionId" = ?',
				#	$sessionId);

				// Find the pageview with that update time. Use the latest if there are multiple.
				#$lastPageviewId = intval($db->QueryValue(
				#	'SELECT MAX("PageviewId") FROM "#pageview" WHERE "PageviewUpdateTime" = ?',
				#	$lastPageviewUpdateTime));
				// Alternative: Always use the last-started pageview
				$lastPageviewId = intval($db->QueryValue(
					'SELECT MAX("PageviewId") FROM "#pageview" WHERE "SessionId" = ?',
					$sessionId));
				
				// Clear exit flag from all pageviews that may have been left
				$db->Execute(
					'UPDATE "#pageview" SET "ExitPage" = NULL WHERE "SessionId" = ?',
					intval($sessionId));
				// Set exit flag on the last pageview
				$db->Execute(
					'UPDATE "#pageview" SET "ExitPage" = 1 WHERE "PageviewId" = ?',
					intval($lastPageviewId));

				// Compute and store total session duration
				$sessionDuration = UnbDatabase::ConvertDateTime($session['SessionUpdateTime']) -
					UnbDatabase::ConvertDateTime($session['SessionStartTime']);
				$db->Execute(
					'UPDATE "#session" SET "SessionDuration" = ? WHERE "SessionId" = ?',
					array($sessionDuration, $sessionId));
			}
			$db->Commit();
		}
		catch (Exception $ex)
		{
			$db->RollBack();
			throw $ex;
		}
	}
	
	// Deletes all pageviews of a certain age which are considered temporary, like with the
	// Prerender or CampaignRedirect flag set. Also delete all sessions that have no pageviews
	// anymore after that.
	//
	public function PurgeTempPageviews()
	{
		// Session expiry time, with a bit safety margin
		$time = time() - (intval(UnbConfiguration::Get('session.maxidle', '30')) + 1) * 60;
		$time = UnbDatabase::ConvertDateTime($time);

		$db = UnbDatabase::GetInstance();
		$db->StartTransaction();
		try
		{
			$db->Execute(
				'DELETE FROM "#pageview"
				WHERE "PageviewTime" < ?
					AND ("CampaignRedirect" OR "Prerender")',
				$time);

			$db->Execute(
				'DELETE FROM "#session"
				WHERE NOT EXISTS (
					SELECT "PageviewId"
					FROM "#pageview" p
					WHERE p."SessionId" = "#session"."SessionId")');

			$db->Commit();
		}
		catch (Exception $ex)
		{
			$db->RollBack();
			throw $ex;
		}
	}


	// -------------------------------  Database: Action functions  -------------------------------
	
	
	// Finds existing actions that are a duplicate of the passed data which could be used to
	// create a new action.
	//
	private function FindDuplicateActions($now, $pageviewId, $type, $url, $data)
	{
		// Read all actions of similar time from the same pageview, but not the passed action
		// (if it already exists in the database)
		$maxOffset = 10;   // allowed offset in seconds
		$db = UnbDatabase::GetInstance();
		$recs = $db->Query(
			'SELECT a.*, t1."Value" as "ActionUrl", t2."Value" as "ActionData"
			FROM "#action" a
				LEFT JOIN "#text" t1 ON (t1."TextId" = a."ActionUrlText")
				LEFT JOIN "#text" t2 ON (t2."TextId" = a."ActionDataText")
			WHERE "PageviewId" = ?
				AND "ActionTime" >= ?
				AND "ActionTime" <= ?
			ORDER BY "ActionId" DESC',
			array(
				$pageviewId,
				UnbDatabase::ConvertDateTime($now - $maxOffset),
				UnbDatabase::ConvertDateTime($now + $maxOffset)
			));
		if (!$recs) return array();
		
		$dupes = array();
		foreach ($recs as $action)
		{
			// We're iterating all actions in the allowed time in reverse order of creation.
			// Collect all duplicates as long as the URL/data matches. As soon as the URL/data is
			// different, stop considering actions as duplicate, because even if we find an
			// older action with the same URL/data, the other intermediate actions indicate that
			// there has been a legitimate navigation meanwhile and the new request must be a
			// different action, i.e. not a duplicate.
			if ($action['ActionType'] == $type &&
				$action['ActionUrl'] == $url &&
				$action['ActionData'] == $data)
			{
				$dupes[] = $action;
			}
			else
			{
				break;
			}
		}
		return $dupes;
	}
	
	// Creates a new action in the database.
	//
	// Returns an array of the action data.
	//
	private function CreateAction($now, $pageviewId, $type, $url, $data)
	{
		$values = array(
			'ActionId' => new UnbDbAutoValue('action'),
			'PageviewId' => $pageviewId,
			'ActionTime' => UnbDatabase::ConvertDateTime($now),
			'ActionType' => $type,
			);

		if (strlen($url))
			$values['ActionUrlText'] = Tomo::GetTextId($url);
		if (strlen($data))
			$values['ActionDataText'] = Tomo::GetTextId($data);

		$db = UnbDatabase::GetInstance();
		$db->StartTransaction();
		try
		{
			$db->Insert('action', $values);

			$db->Commit();
		}
		catch (Exception $ex)
		{
			$db->RollBack();
			throw $ex;
		}

		$values['ActionId'] = $db->GetLastAutoValue();
		return $values;
	}


	// --------------------------------  Database: Text functions  --------------------------------
	
	
	// Gets the text ID for the specified text or inserts it to the database and returns the new
	// text ID. If the text is empty, it will not be added to the dictionary and null is returned
	// because in this application, empty strings are always considered 'not available', just as
	// NULL values.
	//
	// text = (string) Text value to insert and/or get the text ID for.
	//
	// Returns (int) The text ID of the text value, whether it existed before or was just inserted,
	//               or {{null}} if the text is empty.
	//
	public static function GetTextId($text)
	{
		$text = trim($text);
		if (!strlen($text)) return null;

		$db = UnbDatabase::GetInstance();
		$textId = $db->QueryValue('SELECT "TextId" FROM "#text" WHERE "Value" = ?', $text);
		if (!$textId)
		{
			$db->StartTransaction();
			try
			{
				$db->Execute(
					'INSERT INTO "#text" ("TextId", "Value") VALUES (?, ?)',
					array(new UnbDbAutoValue('text'), $text));

				$textId = $db->GetLastAutoValue();
				$db->Commit();
			}
			catch (Exception $ex)
			{
				$db->RollBack();
				throw $ex;
			}
		}
		return $textId;
	}
	
	// Finds a text string by its text ID.
	//
	// textId = (int) Text ID
	//
	// Throws InvalidArgumentException on invalid argument type or value.
	// Throws UnbDatabaseException
	//
	// Returns the text string if successful, {{null}} otherwise.
	//
	public static function FindTextById($textId)
	{
		if (!is_int($textId))
			throw new InvalidArgumentException('Invalid argument type: $textId');
		if ($textId <= 0)
			throw new InvalidArgumentException('Argument out of range: $textId');

		$db = UnbDatabase::GetInstance();
		$text = $db->QueryValue('SELECT "Value" FROM "#text" WHERE "TextId" = ?', $textId);
		return $text;
	}

	// Purges unreferenced text data and frees up spare space in the database.
	//
	public static function PurgeText()
	{
		$sessionVars = '';
		foreach (self::GetDbSessionVars() as $name => $type)
		{
			if ($type == 'text')
			{
				$sessionVars .= ' OR "sv_' . $name . '" = "TextId"';
			}
		}
		$pageviewVars = '';
		foreach (self::GetDbPageviewVars() as $name => $type)
		{
			if ($type == 'text')
			{
				$pageviewVars .= ' OR "pv_' . $name . '" = "TextId"';
			}
		}
		
		$db = UnbDatabase::GetInstance();
		$db->StartTransaction();
		try
		{
			foreach ($db->QueryArray(
				'SELECT "TextId"
				FROM "#text"
				WHERE NOT EXISTS (
						SELECT 1
						FROM "#session"
						WHERE "ReferrerHostText" = "TextId" OR
							"ReferrerUrlPathText" = "TextId" OR
							"SearchEngineText" = "TextId" OR
							"SearchKeywordsText" = "TextId" OR
							"SearchImageUrlText" = "TextId" OR
							"CampaignText" = "TextId" OR
							"ProviderText" = "TextId" OR
							"CityText" = "TextId" OR
							"RegionText" = "TextId" OR
							"BrowserText" = "TextId" OR
							"OperatingSystemText" = "TextId"' . $sessionVars . ') AND
					NOT EXISTS (
						SELECT 1
						FROM "#pageview"
						WHERE "DomainText" = "TextId" OR
							"UrlPathText" = "TextId" OR
							"UrlQueryText" = "TextId" OR
							"UrlHashText" = "TextId" OR
							"PageTitleText" = "TextId"' . $pageviewVars . ') AND
					NOT EXISTS (
						SELECT 1
						FROM "#action"
						WHERE "ActionUrlText" = "TextId" OR
							"ActionDataText" = "TextId")') as $textId)
			{
				$textId = intval($textId);
				$db->Execute(
					'DELETE FROM "#text" WHERE "TextId" = ?',
					array($textId));
			}

			$db->Commit();
		}
		catch (Exception $ex)
		{
			$db->RollBack();
			throw $ex;
		}

		$db->Optimise(true);
	}


	// ------------------------------  Database: Settings functions  ------------------------------
	
	
	public static function GetSetting($name, $defaultValue = null)
	{
		$db = UnbDatabase::GetInstance();
		$value = $db->QueryValue('SELECT "Value" FROM "#setting" WHERE "Name" = ?', $name);
		if (!isset($value))
		{
			$value = $defaultValue;
		}
		return $value;
	}
	
	public static function SetSetting($name, $value)
	{
		$db = UnbDatabase::GetInstance();
		$db->Update('setting', array('Value' => $value), '"Name" = ?', $name);
		if (!$db->GetRowsAffected())
		{
			try
			{
				$db->Insert('setting', array('Name' => $name, 'Value' => $value));
			}
			catch (Exception $ex)
			{
			}
		}
	}
	

	// --------------------------------  Database: Lock functions  --------------------------------
	
	
	// Tries to acquire a lock on the remote IP address. If the lock is acquired successfully, no
	// other lock-guarded operation can be performed for the same remote IP address. If the lock
	// cannot be acquired immediately, false is returned. It is the responsibility of the caller
	// to retry.
	//
	// Returns true if the lock was acquired, false otherwise.
	//
	public function AcquireLock()
	{
		$id = UnbEnvironment::GetClientAddress();

		// First purge all stale locks, should there be any
		$this->PurgeLocks();
		
		$db = UnbDatabase::GetInstance();
		$db->StartTransaction();
		try
		{
			$values = array(
				'LockId' => $id,
				'LockCreateTime' => UnbDatabase::ConvertDateTime(time())
				);
			
			$db->Insert('lock', $values);

			$db->Commit();
		}
		catch (UnbDatabaseException $ex)
		{
			if ($db->IsConstraintError())
			{
				// SQLite doesn't say something about "PRIMARY"...
				#if (strpos($db->GetErrorMessage(), 'PRIMARY') !== false)
				{
					$db->RollBack();
					return false;
				}
			}
			$db->RollBack();
			throw $ex;
		}
		catch (Exception $ex)
		{
			$db->RollBack();
			throw $ex;
		}
		$this->lockId = $id;
		return true;
	}
	
	// Releases a previously acquired lock on the remote IP address, if any.
	//
	public function ReleaseLock()
	{
		if (!isset($this->lockId)) return;
		UnbDebug::Trace('ReleaseLock: deleting id: ' . $this->lockId);

		$db = UnbDatabase::GetInstance();
		$db->StartTransaction();
		try
		{
			$db->Delete('lock', 'LockId = ?', $this->lockId);
			$db->Commit();

			$count = $db->QueryValue('SELECT count(*) from "#lock"');
			if ($count) UnbDebug::Trace('ReleaseLock: remaining record count: ' . $count);
		}
		catch (Exception $ex)
		{
			$db->RollBack();
			throw $ex;
		}
		$this->lockId = null;
	}
	
	// Purges all locks that have been acquired a long time ago and are very likely stale.
	//
	public function PurgeLocks()
	{
		// Maximum allowed lock lifetime in seconds
		$lockTimeout = 15;
		
		$db = UnbDatabase::GetInstance();
		$db->StartTransaction();
		try
		{
			$db->Delete('lock', 'LockCreateTime < ?', UnbDatabase::ConvertDateTime(time() - $lockTimeout));
			$db->Commit();
		}
		catch (Exception $ex)
		{
			$db->RollBack();
			throw $ex;
		}
	}


	// --------------------------  Database: Custom variables management  -------------------------


	// Gets all configured custom visitor variables.
	//
	public static function GetConfVisitorVars()
	{
		$confVisitorVars = UnbConfiguration::Get('custom-vars.visitor');
		if (!is_array($confVisitorVars))
			$confVisitorVars = array();
		return $confVisitorVars;
	}
	
	// Gets all configured custom session variables.
	//
	public static function GetConfSessionVars()
	{
		$confSessionVars = UnbConfiguration::Get('custom-vars.session');
		if (!is_array($confSessionVars))
			$confSessionVars = array();
		return $confSessionVars;
	}
	
	// Gets all configured custom pageview variables.
	//
	public static function GetConfPageviewVars()
	{
		$confPageviewVars = UnbConfiguration::Get('custom-vars.pageview');
		if (!is_array($confPageviewVars))
			$confPageviewVars = array();
		return $confPageviewVars;
	}
	
	// Gets all custom visitor variables from the database table columns.
	//
	public static function GetDbVisitorVars()
	{
		return self::GetDbCustomVars('visitor');
	}
	
	// Gets all custom session variables from the database table columns.
	//
	public static function GetDbSessionVars()
	{
		return self::GetDbCustomVars('session');
	}
	
	// Gets all custom pageview variables from the database table columns.
	//
	public static function GetDbPageviewVars()
	{
		return self::GetDbCustomVars('pageview');
	}
	
	// Gets all custom variables in a scope from the database table columns.
	//
	// scope = (string) "visitor", "session" or "pageview"
	//
	private static function GetDbCustomVars($scope)
	{
		$db = UnbDatabase::GetInstance();
		$columns = $db->Query('PRAGMA table_info("#' . $scope . '")');
		$tableSql = $db->QueryValue('SELECT sql FROM sqlite_master WHERE type = \'table\' AND name LIKE \'%' . $scope . '\'');
		$dbVars = array();
		foreach ($columns as $col)
		{
			if (preg_match('/^' . substr($scope, 0, 1) . 'v_(.+)$/', $col['name'], $m))
			{
				if (preg_match('/"' . $col['name'] . '" INTEGER REFERENCES ".*text" \("TextId"\)/i', $tableSql))
					$type = 'text';
				else if (preg_match('/"' . $col['name'] . '" INTEGER/i', $tableSql))
					$type = 'integer';
				else
					$type = 'unknown';
				$dbVars[$m[1]] = $type;
			}
		}
		return $dbVars;
	}
	
	// Adds a custom visitor variable as new table column into the database.
	//
	// name = (string) Name of the new variable
	//
	// The variable type is looked up in the configuration. If the variable does not exist in the
	// configuration, it will not be created.
	//
	// Throws UnbDatabaseException if the column could not be added to the table.
	//
	public function AddVisitorVar($name)
	{
		$this->AddCustomVar('visitor', $name);
	}
	
	// Adds a custom session variable as new table column into the database.
	//
	// name = (string) Name of the new variable
	//
	// The variable type is looked up in the configuration. If the variable does not exist in the
	// configuration, it will not be created.
	//
	// Throws UnbDatabaseException if the column could not be added to the table.
	//
	public function AddSessionVar($name)
	{
		$this->AddCustomVar('session', $name);
	}
	
	// Adds a custom pageview variable as new table column into the database.
	//
	// name = (string) Name of the new variable
	//
	// The variable type is looked up in the configuration. If the variable does not exist in the
	// configuration, it will not be created.
	//
	// Throws UnbDatabaseException if the column could not be added to the table.
	//
	public function AddPageviewVar($name)
	{
		$this->AddCustomVar('pageview', $name);
	}
	
	// Adds a custom variable as new table column into the database.
	//
	// scope = (string) "visitor", "session" or "pageview"
	// name = (string) Name of the new variable
	//
	// The variable type is looked up in the configuration. If the variable does not exist in the
	// configuration, it will not be created.
	//
	// Throws UnbDatabaseException if the column could not be added to the table.
	//
	public function AddCustomVar($scope, $name)
	{
		if (preg_match(self::varNamePattern, $name))
		{
			$customVars = UnbConfiguration::Get('custom-vars.' . $scope);
			$type = $customVars[$name];
			switch ($type)
			{
				case 'integer':
					$type = 'INTEGER';
					break;
				case 'text':
					$type = 'INTEGER REFERENCES "#text" ("TextId")';
					break;
				default:
					$type = null;
			}
			if (isset($type))
			{
				$db = UnbDatabase::GetInstance();
				$db->Execute('ALTER TABLE "#' . $scope . '" ADD COLUMN "' . substr($scope, 0, 1) . 'v_' . $name . '" ' . $type);
			}
		}
	}


	// ---------------------------  Database: Custom action management  ---------------------------


	// Gets all configured custom actions.
	//
	public static function GetConfActions()
	{
		$confActions = UnbConfiguration::Get('custom-actions');
		if (!is_array($confActions))
			$confActions = array();
		return $confActions;
	}
	

	// ----------------------------------  Database: Read access  ---------------------------------


	public function GetAllColumnNames()
	{
		$cols = array(
			'ActionData',
			'ActionId',
			'ActionTime',
			'ActionType',
			'ActionUrl',
			'ActivityTimespan',
			'AdBlock',
			'Browser',
			'Campaign',
			'CampaignRedirect',
			'City',
			'Cookies',
			'CountryCode',
			'DNT',
			'Domain',
			'EntryPage',
			'ExitPage',
			'Flash',
			'IPAddress',
			'IPVersion',
			'Java',
			'JavaScript',
			'JavaScriptData',
			'Language',
			'Lat',
			'Lng',
			'Mobile',
			'NewVisitor',
			'OperatingSystem',
			'OriginalPageviewId',
			'PageLoadTime',
			'PageTitle',
			'PageLanguage',
			'PageviewId',
			'PageviewTime',
			'PageviewUpdateTime',
			'Prerender',
			'Provider',
			'Quicktime',
			'ReferrerHost',
			'ReferrerUrlPath',
			'ReferrerSsl',
			'Region',
			'ScreenAvailHeight',
			'ScreenAvailWidth',
			'ScreenColors',
			'ScreenHeight',
			'ScreenWidth',
			'SearchEngine',
			'SearchImageUrl',
			'SearchKeywords',
			'SearchResultPage',
			'SessionDuration',
			'SessionId',
			'SessionStartTime',
			'SessionUpdateTime',
			'Silverlight',
			'Ssl',
			'UrlHash',
			'UrlPath',
			'UrlQuery',
			'UtcOffset',
			'ViewportHeight',
			'ViewportWidth',
			'VisitorId',
			'WindowsMedia',
		);
		
		foreach (self::GetDbVisitorVars() as $name => $type)
		{
			$cols[] = 'vv_' . $name;
		}
		foreach (self::GetDbSessionVars() as $name => $type)
		{
			$cols[] = 'sv_' . $name;
		}
		foreach (self::GetDbPageviewVars() as $name => $type)
		{
			$cols[] = 'pv_' . $name;
		}
		
		return $cols;
	}

	public function GetPageviews($columns, $filterData, $reverse = false)
	{
		$dbVisitorVars = self::GetDbVisitorVars();
		$dbSessionVars = self::GetDbSessionVars();
		$dbPageviewVars = self::GetDbPageviewVars();

		$cols = '';
		$joins = '';
		$where = '';
		$whereParams = array();
		$needVisitorTable = false;
		foreach ($columns as $col)
		{
			switch (strtolower($col))
			{
				case 'referrerhost':
				case 'referrerurlpath':
				case 'city':
				case 'region':
				case 'searchengine':
				case 'searchimageurl':
				case 'searchkeywords':
				case 'campaign':
				case 'provider':
				case 'browser':
				case 'operatingsystem':
					$cols .= ($cols ? ', ' : '') .
						't_' . $col . '."Value" "' . $col . '"';
					$joins .= 'LEFT JOIN "#text" t_' . $col . ' ON (t_' . $col . '."TextId" = "#session"."' . $col . 'Text") ';
					break;
				case 'domain':
				case 'urlpath':
				case 'urlquery':
				case 'urlhash':
				case 'pagetitle':
					$cols .= ($cols ? ', ' : '') .
						't_' . $col . '."Value" "' . $col . '"';
					$joins .= 'LEFT JOIN "#text" t_' . $col . ' ON (t_' . $col . '."TextId" = "#pageview"."' . $col . 'Text") ';
					break;
				case 'actiondata':
				case 'actionurl':
					$cols .= ($cols ? ', ' : '') .
						't_' . $col . '."Value" "' . $col . '"';
					$joins .= 'LEFT JOIN "#text" t_' . $col . ' ON (t_' . $col . '."TextId" = "#action"."' . $col . 'Text") ';
					break;
				case 'visitorid':
					// VisitorId is "abiguous" otherwise, no idea why.
					$cols .= ($cols ? ', ' : '') .
						'"#session"."VisitorId"';
					break;
				default:
					if (substr($col, 0, 3) == 'vv_' && $dbVisitorVars[substr($col, 3)] == 'text')
					{
						$cols .= ($cols ? ', ' : '') .
							't_' . $col . '."Value" "' . $col . '"';
						$joins .= 'LEFT JOIN "#text" t_' . $col . ' ON (t_' . $col . '."TextId" = "#visitor"."' . $col . '") ';
					}
					else if (substr($col, 0, 3) == 'sv_' && $dbSessionVars[substr($col, 3)] == 'text')
					{
						$cols .= ($cols ? ', ' : '') .
							't_' . $col . '."Value" "' . $col . '"';
						$joins .= 'LEFT JOIN "#text" t_' . $col . ' ON (t_' . $col . '."TextId" = "#session"."' . $col . '") ';
					}
					else if (substr($col, 0, 3) == 'pv_' && $dbPageviewVars[substr($col, 3)] == 'text')
					{
						$cols .= ($cols ? ', ' : '') .
							't_' . $col . '."Value" "' . $col . '"';
						$joins .= 'LEFT JOIN "#text" t_' . $col . ' ON (t_' . $col . '."TextId" = "#pageview"."' . $col . '") ';
					}
					else
					{
						$cols .= ($cols ? ', ' : '') .
							'"' . $col . '"';
					}
					
					if (substr($col, 0, 3) == 'vv_')
					{
						$needVisitorTable = true;
					}
					break;
			}
		}

		if ($filterData['filtercol'])
		{
			$filterWhere = '';
			foreach ($filterData['filtercol'] as $col)
			{
				$active = $filterData[$col . '_active'];
				if (!$active) continue;

				$realCol = '"' . $col . '"';
				if ($col == 'VisitorId')
				{
					// VisitorId is "abiguous" otherwise, no idea why. (See above)
					$realCol = '"#session"."VisitorId"';
				}
				if (substr($col, 0, 3) == 'vv_' && $dbVisitorVars[substr($col, 3)] == 'text' ||
					substr($col, 0, 3) == 'sv_' && $dbSessionVars[substr($col, 3)] == 'text' ||
					substr($col, 0, 3) == 'pv_' && $dbPageviewVars[substr($col, 3)] == 'text')
				{
					// Custom variables text columns need to be specified as in the source column.
					// They don't have that *Text name mapping. When specifiying the raw column
					// name, SQLite might try to compare the original table column (which only
					// contains the TextId) instead of the joined text string (I assume). At least
					// this is what works.
					$realCol = 't_' . $col . '."Value"';
				}
				
				$op = $filterData[$col . '_op'];
				$value = $filterData[$col . '_value'];
				$value2 = $filterData[$col . '_value2'];
				$not = $filterData[$col . '_not'];

				$filterWhere .= ($filterWhere ? ' ' . ($col['allfilters'] ? 'AND' : 'OR') . ' ' : '');

				if ($not)
				{
					$filterWhere .= 'NOT ';
				}
				$filterWhere .= '(';
				switch ($op)
				{
					case 'lt':
						$filterWhere .= $realCol . ' < ?';
						$whereParams[] = $value;
						break;
					case 'lte':
						$filterWhere .= $realCol . ' <= ?';
						$whereParams[] = $value;
						break;
					case 'gt':
						$filterWhere .= $realCol . ' > ?';
						$whereParams[] = $value;
						break;
					case 'gte':
						$filterWhere .= $realCol . ' >= ?';
						$whereParams[] = $value;
						break;
					case 'ne':
						$filterWhere .= $realCol . ' <> ? OR ' . $realCol . ' IS NULL';
						$whereParams[] = $value;
						break;
					case 'between':
						$filterWhere .= $realCol . ' BETWEEN ? AND ?';
						$whereParams[] = $value;
						$whereParams[] = $value2;
						break;
					case 'startswith':
						$filterWhere .= $realCol . ' LIKE ?';
						$whereParams[] = $value . '%';
						break;
					case 'endswith':
						$filterWhere .= $realCol . ' LIKE ?';
						$whereParams[] = '%' . $value;
						break;
					case 'contains':
						$filterWhere .= $realCol . ' LIKE ?';
						$whereParams[] = '%' . $value . '%';
						break;
					case 'isnull':
						$filterWhere .= $realCol . ' IS NULL';
						break;
					case 'eq':
					default:
						$filterWhere .= $realCol . ' = ?';
						$whereParams[] = $value;
						break;
				}
				$filterWhere .= ')';
			}
			if (strlen($filterWhere))
			{
				$where .= ($where ? ' AND ' : '') .
					'(' . $filterWhere . ')';
			}
		}

		if ($needVisitorTable)
		{
			$joins = 'JOIN "#visitor" USING ("VisitorId") ' . $joins;
		}
		
		$sql = 'SELECT ' . $cols . ' FROM "#session" LEFT JOIN "#pageview" USING ("SessionId") LEFT JOIN "#action" USING ("PageviewId") ' . $joins;
		$params = array();

		if ($where)
		{
			$sql .= ' WHERE ' . $where;
			foreach ($whereParams as $p)
			{
				$params[] = $p;
			}
		}

		$revStr = $reverse ? ' DESC' : '';
		$orderArray = array('"SessionId"' . $revStr, '"PageviewId"' . $revStr, '"ActionId"' . $revStr);
		if (count($orderArray))
		{
			#ksort($orderArray);
			$sql .= ' ORDER BY ' . join(', ', $orderArray);
		}

		$db = UnbDatabase::GetInstance();
		return $db->Query($sql, $params);
	}

	public function GetRecords($withPageviews, $withActions, $columns, $limits, $groupLevel = 1, $groupValues = null)
	{
		if (!count($columns)) return array();

		#echo '<pre>groupValues = ';
		#print_r($groupValues);
		#echo '</pre>';

		$dbVisitorVars = self::GetDbVisitorVars();
		$dbSessionVars = self::GetDbSessionVars();
		$dbPageviewVars = self::GetDbPageviewVars();

		$cols = '';
		$joins = '';
		$where = '';
		$whereParams = array();
		$having = '';
		$havingParams = array();
		$group = '';
		$orderArray = array();
		$maxGroup = 0;
		#$groupLevel = $groupValues ? count($groupValues) + 1 : 1;
		$needVisitorTable = false;

		foreach ($columns as $col)
		{
			if (!strlen($col['source'])) continue;   // Not a source column

			$colRealName = '';
			$maxGroup = max($maxGroup, $col['group']);

			switch ($col['source'])
			{
				case 'ReferrerHost':
				case 'ReferrerUrlPath':
				case 'SearchEngine':
				case 'SearchImageUrl':
				case 'SearchKeywords':
				case 'Campaign':
				case 'Provider':
				case 'City':
				case 'Region':
				case 'Browser':
				case 'OperatingSystem':
					if ($col['group'] <= $groupLevel)
					{
						$cols .= ($cols ? ', ' : '') .
							't_' . $col['source'] . '."Value" "' . $col['alias'] . '"';
					}
					$colRealName = 't_' . $col['source'] . '."Value"';
					$joins .= ' LEFT JOIN "#text" t_' . $col['source'] . ' ON (t_' . $col['source'] . '."TextId" = "#session"."' . $col['source'] . 'Text")';
					break;
				case 'Domain':
				case 'UrlPath':
				case 'UrlQuery':
				case 'UrlHash':
				case 'PageTitle':
					if ($col['group'] <= $groupLevel)
					{
						$cols .= ($cols ? ', ' : '') .
							't_' . $col['source'] . '."Value" "' . $col['alias'] . '"';
					}
					$colRealName = 't_' . $col['source'] . '."Value"';
					$joins .= ' LEFT JOIN "#text" t_' . $col['source'] . ' ON (t_' . $col['source'] . '."TextId" = "#pageview"."' . $col['source'] . 'Text")';
					break;
				case 'ActionUrl':
				case 'ActionData':
					if ($col['group'] <= $groupLevel)
					{
						$cols .= ($cols ? ', ' : '') .
							't_' . $col['source'] . '."Value" "' . $col['alias'] . '"';
					}
					$colRealName = 't_' . $col['source'] . '."Value"';
					$joins .= ' LEFT JOIN "#text" t_' . $col['source'] . ' ON (t_' . $col['source'] . '."TextId" = "#action"."' . $col['source'] . 'Text")';
					break;
				default:
					if ($col['function'])
					{
						if ($col['group'] <= $groupLevel)
						{
							$cols .= ($cols ? ', ' : '') .
								$col['source'] . ' "' . $col['alias'] . '"';
						}
						$colRealName = '"' . $col['alias'] . '"';
					}
					else
					{
						if ($col['group'] <= $groupLevel)
						{
							if (substr($col['source'], 0, 3) == 'vv_' && $dbVisitorVars[substr($col['source'], 3)] == 'text')
							{
								$cols .= ($cols ? ', ' : '') .
									't_' . $col['source'] . '."Value" "' . $col['alias'] . '"';
								$colRealName = 't_' . $col['source'] . '."Value"';
								$joins .= ' LEFT JOIN "#text" t_' . $col['source'] . ' ON (t_' . $col['source'] . '."TextId" = "#visitor"."' . $col['source'] . '")';
							}
							else if (substr($col['source'], 0, 3) == 'sv_' && $dbSessionVars[substr($col['source'], 3)] == 'text')
							{
								$cols .= ($cols ? ', ' : '') .
									't_' . $col['source'] . '."Value" "' . $col['alias'] . '"';
								$colRealName = 't_' . $col['source'] . '."Value"';
								$joins .= ' LEFT JOIN "#text" t_' . $col['source'] . ' ON (t_' . $col['source'] . '."TextId" = "#session"."' . $col['source'] . '")';
							}
							else if (substr($col['source'], 0, 3) == 'pv_' && $dbPageviewVars[substr($col['source'], 3)] == 'text')
							{
								$cols .= ($cols ? ', ' : '') .
									't_' . $col['source'] . '."Value" "' . $col['alias'] . '"';
								$colRealName = 't_' . $col['source'] . '."Value"';
								$joins .= ' LEFT JOIN "#text" t_' . $col['source'] . ' ON (t_' . $col['source'] . '."TextId" = "#pageview"."' . $col['source'] . '")';
							}
							else
							{
								$cols .= ($cols ? ', ' : '') .
									'"' . $col['source'] . '" "' . $col['alias'] . '"';
							}

							if (substr($col['source'], 0, 3) == 'vv_')
							{
								$needVisitorTable = true;
							}
						}
						$colRealName = '"' . $col['source'] . '"';
					}
					break;
			}

			if ($col['set'] === true)
			{
				$where .= ($where ? ' AND ' : '') .
					'(' . $colRealName . ' IS NOT NULL AND ' . $colRealName . ' <> \'\')';
			}
			else if ($col['set'] === false)
			{
				$where .= ($where ? ' AND ' : '') .
					'(' . $colRealName . ' IS NULL OR ' . $colRealName . ' = \'\')';
			}

			if (is_array($col['filters']) && count($col['filters']))
			{
				$filterWhere = '';
				foreach ($col['filters'] as $filter)
				{
					$filterWhere .= ($filterWhere ? ' ' . ($col['allfilters'] ? 'AND' : 'OR') . ' ' : '');

					if ($filter['not'])
					{
						$filterWhere .= 'NOT ';
					}
					$filterWhere .= '(';
					switch ($filter['op'])
					{
						case '<':
						case 'lt':
							$filterWhere .= $colRealName . ' < ?';
							$whereParams[] = $filter['value'];
							break;
						case '<=':
						case 'lte':
							$filterWhere .= $colRealName . ' <= ?';
							$whereParams[] = $filter['value'];
							break;
						case '>':
						case 'gt':
							$filterWhere .= $colRealName . ' > ?';
							$whereParams[] = $filter['value'];
							break;
						case '>=':
						case 'gte':
							$filterWhere .= $colRealName . ' >= ?';
							$whereParams[] = $filter['value'];
							break;
						case '<>':
						case '!=':
						case 'ne':
							$filterWhere .= $colRealName . ' <> ? OR ' . $colRealName . ' IS NULL';
							$whereParams[] = $filter['value'];
							break;
						case 'between':
							$filterWhere .= $colRealName . ' BETWEEN ? AND ?';
							$whereParams[] = $filter['value'];
							$whereParams[] = $filter['value2'];
							break;
						case 'startswith':
							$filterWhere .= $colRealName . ' LIKE ?';
							$whereParams[] = $filter['value'] . '%';
							break;
						case 'endswith':
							$filterWhere .= $colRealName . ' LIKE ?';
							$whereParams[] = '%' . $filter['value'];
							break;
						case 'contains':
							$filterWhere .= $colRealName . ' LIKE ?';
							$whereParams[] = '%' . $filter['value'] . '%';
							break;
						case '=':
						case 'eq':
						default:
							$filterWhere .= $colRealName . ' = ?';
							$whereParams[] = $filter['value'];
							break;
					}
					$filterWhere .= ')';
				}
				$where .= ($where ? ' AND ' : '') .
					'(' . $filterWhere . ')';
			}
			if (is_array($col['pgfilters']) && count($col['pgfilters']))
			{
				$filterHaving = '';
				foreach ($col['pgfilters'] as $filter)
				{
					$filterHaving .= ($filterHaving ? ' ' . ($col['allpgfilters'] ? 'AND' : 'OR') . ' ' : '');

					if ($filter['not'])
					{
						$filterHaving .= 'NOT ';
					}
					$filterHaving .= '(';
					switch ($filter['op'])
					{
						case '<':
						case 'lt':
							$filterHaving .= $colRealName . ' < ?';
							$havingParams[] = $filter['value'];
							break;
						case '<=':
						case 'lte':
							$filterHaving .= $colRealName . ' <= ?';
							$havingParams[] = $filter['value'];
							break;
						case '>':
						case 'gt':
							$filterHaving .= $colRealName . ' > ?';
							$havingParams[] = $filter['value'];
							break;
						case '>=':
						case 'gte':
							$filterHaving .= $colRealName . ' >= ?';
							$havingParams[] = $filter['value'];
							break;
						case '<>':
						case '!=':
						case 'ne':
							$filterHaving .= $colRealName . ' <> ? OR ' . $colRealName . ' IS NULL';
							$havingParams[] = $filter['value'];
							break;
						case 'between':
							$filterHaving .= $colRealName . ' BETWEEN ? AND ?';
							$havingParams[] = $filter['value'];
							$havingParams[] = $filter['value2'];
							break;
						case 'startswith':
							$filterHaving .= $colRealName . ' LIKE ?';
							$havingParams[] = $filter['value'] . '%';
							break;
						case 'endswith':
							$filterHaving .= $colRealName . ' LIKE ?';
							$havingParams[] = '%' . $filter['value'];
							break;
						case 'contains':
							$filterHaving .= $colRealName . ' LIKE ?';
							$havingParams[] = '%' . $filter['value'] . '%';
							break;
						case '=':
						case 'eq':
						default:
							$filterHaving .= $colRealName . ' = ?';
							$havingParams[] = $filter['value'];
							break;
					}
					$filterHaving .= ')';
				}
				$having .= ($having ? ' AND ' : '') .
					'(' . $filterHaving . ')';
			}

			if ($col['group'] <= $groupLevel)
			{
				if ($col['group'])
				{
					$group .= ($group ? ', ' : '') .
						'"' . $col['alias']. '"';
				}
				if ($col['group'] && $col['group'] < $groupLevel)
				{
					$groupValue = $groupValues[$col['alias']];
					if (isset($groupValue))
					{
						$where .= ($where ? ' AND ' : '') .
							'("' . $col['alias']. '" = ?)';
						$whereParams[] = $groupValue;
					}
					else
					{
						$where .= ($where ? ' AND ' : '') .
							'("' . $col['alias']. '" IS NULL)';
					}
				}

				if ($col['sortindex'])
				{
					$orderArray[$col['sortindex']] = '"' . $col['alias'] . '"' . ($col['sortdir'] == 'desc' ? ' DESC' : '');
				}
			}
		}

		$sql = 'SELECT ' . $cols . ' FROM "#session"';
		if ($needVisitorTable)
		{
			$sql .= ' JOIN "#visitor" USING ("VisitorId")';
		}
		if ($withPageviews || $withActions)
		{
			$sql .= ' LEFT JOIN "#pageview" USING ("SessionId")';
		}
		if ($withActions)
		{
			$sql .= ' LEFT JOIN "#action" USING ("PageviewId")';
		}
		
		$sql .= $joins;
		$params = array();

		if ($where)
		{
			$sql .= ' WHERE ' . $where;
			foreach ($whereParams as $p)
			{
				$params[] = $p;
			}
		}

		if ($group)
		{
			$sql .= ' GROUP BY ' . $group;
		}

		if ($having)
		{
			$sql .= ' HAVING ' . $having;
			foreach ($havingParams as $p)
			{
				$params[] = $p;
			}
		}

		if (count($orderArray))
		{
			ksort($orderArray);
			$sql .= ' ORDER BY ' . join(', ', $orderArray);
		}

		if ($limits[$groupLevel - 1])
		{
			$sql .= ' LIMIT ' . $limits[$groupLevel - 1];
		}

		$db = UnbDatabase::GetInstance();
		$recs = $db->Query($sql, $params);
		
		if (!$recs)
		{
			$db->traceSql = true;
			$db->Query($sql, $params);
			$db->traceSql = false;
		}

		$outRecs = array();
		foreach ($recs as $rec)
		{
			$rec['_group'] = $groupLevel;

			$outRecs[] = $rec;

			if ($groupLevel < $maxGroup)
			{
				// There are deeper grouping levels, go find them
				$subGroupValues = array();
				foreach ($columns as $col)
				{
					if ($col['group'] && $col['group'] <= $groupLevel)
					{
						$subGroupValues[$col['alias']] = $rec[$col['alias']];
					}
				}
				foreach ($this->GetRecords($withPageviews, $withActions, $columns, $limits, $groupLevel + 1, $subGroupValues) as $subRec)
				{
					$outRecs[] = $subRec;
				}
			}
		}
		return $outRecs;
	}
	
	// Gets the start time of the first session, in string format.
	//
	function GetFirstSessionTime()
	{
		$db = UnbDatabase::GetInstance();
		return $db->QueryValue('SELECT MIN("SessionStartTime") FROM "#session"');
	}

} // class Tomo

?>
